# pylint: disable=no-member,redefined-outer-name,unused-argument
from warnings import warn

from django.urls.conf import include, re_path
from django.conf import settings
from rest_framework import routers, permissions, versioning, schemas

from . import responses
from ..utils import import_class


class _AbstractRouter(routers.DefaultRouter):

    def __init__(self, *args, **kwargs):
        self.custom_urls = []
        self.permission_classes = kwargs.pop("perms", None)
        super().__init__(*args, **kwargs)

    def _get_api_root_dict(self):
        return {
            reg[0]: self.routes[0].name.format(basename=reg[2])
            for reg in self.registry
        }

    def _get_custom_lists(self):
        return self.custom_urls

    def _get_views_custom_list(self, view_request, registers):
        routers_list = {}
        fpath = view_request.get_full_path().split("?")
        absolute_uri = view_request.build_absolute_uri(fpath[0])
        for prefix, _, name in self._get_custom_lists():
            path = ''.join([
                absolute_uri, prefix, f'?{fpath[1]}' if len(fpath) > 1 else ""
            ])
            routers_list[name] = path
        routers_list.update(registers.data)
        return routers_list

    def get_default_basename(self, viewset):
        base_name = getattr(viewset, 'base_name', None)
        if base_name is not None:
            return base_name
        queryset = getattr(viewset, 'queryset', None)
        model = getattr(viewset, 'model', None)
        if queryset is None:  # nocv
            assert model is not None, \
                '`base_name` argument not specified, and could ' \
                'not automatically determine the name from the viewset, as ' \
                'it does not have a `.queryset` or `.model` attribute.'
            return model._meta.object_name.lower()
        # can't be tested because this initialization takes place before any
        # test code can be run
        return super().get_default_basename(viewset)

    def register_view(self, prefix, view, name=None):
        if getattr(view, 'as_view', None):
            name = name or view().get_view_name().lower()
            self.custom_urls.append((prefix, view, name))
            return
        name = name or view.get_view_name().lower()
        self.custom_urls.append((prefix, view, name))

    def _unreg(self, prefix, objects_list):
        del self._urls
        index = 0
        for reg_prefix, _, _ in objects_list:
            if reg_prefix == prefix:
                del objects_list[index]
                break
            index += 1  # nocv
        return objects_list

    def unregister_view(self, prefix):
        self.custom_urls = self._unreg(prefix, self.custom_urls)  # nocv

    def unregister(self, prefix):
        self.registry = self._unreg(prefix, self.registry)

    def generate(self, views_list):
        for prefix, options in views_list.items():
            if 'view' in options:
                view = import_class(options['view'])
            elif 'model' in options:
                view = import_class(options['model']).generated_view
            args = [prefix, view]
            if 'name' in options:
                args.append(options['name'])
            elif getattr(view, 'base_name', None) is None:
                args.append(prefix)
            view_type = options.get('type', 'viewset')
            if view_type == 'viewset':
                self.register(*args)
            elif view_type == 'view':  # nocv
                self.register_view(*args)
            else:  # nocv
                raise Exception('Unknown type of view')


class APIRouter(_AbstractRouter):
    default_root_view_name = 'v1'

    def __init__(self, *args, **kwargs):
        self.router_version_name = kwargs.pop('version', self.default_root_view_name)
        self.root_view_name = self.router_version_name
        super().__init__(*args, **kwargs)
        self.__register_schema()
        self.__register_openapi()

    def __register_openapi(self):
        schema_view = import_class(settings.OPENAPI_VIEW_CLASS)
        cache_timeout = settings.SCHEMA_CACHE_TIMEOUT
        swagger_kwargs = {'cache_timeout': 0 if settings.DEBUG else cache_timeout}
        self.register_view(
            '_openapi',
            schema_view.with_ui('swagger', **swagger_kwargs),
            name='_openapi'
        )

    def __register_schema(self, name='schema'):
        try:
            self.register_view(rf'{name}', self._get_schema_view(), name)
        except BaseException as exc:  # nocv
            warn(f"Couldn't attach schema view: {exc}")

    def _get_schema_view(self):
        return schemas.get_schema_view(
            title=self.router_version_name,
            version=self.router_version_name
        )

    def get_api_root_view(self, *args, **kwargs):

        class API(self.APIRootView):
            root_view_name = self.router_version_name
            if self.permission_classes:
                permission_classes = self.permission_classes
            custom_urls = self.custom_urls

            class versioning_class(versioning.BaseVersioning):
                # pylint: disable=invalid-name,no-self-argument

                def determine_version(self_ver, request, *args, **kwargs):
                    return self.router_version_name

            def get_view_name(self): return self.root_view_name

            def get(self_inner, request, *args, **kwargs):
                # pylint: disable=no-self-argument,protected-access
                data = self._get_views_custom_list(
                    request, super(API, self_inner).get(request, *args, **kwargs)
                )
                return responses.HTTP_200_OK(data)

        return API.as_view(api_root_dict=self._get_api_root_dict())

    def get_urls(self):
        urls = super().get_urls()
        for prefix, view, _ in self.custom_urls:
            # pylint: disable=cell-var-from-loop
            view = getattr(view, 'as_view', lambda: view)()
            urls.append(re_path(f"^{prefix}/$", view))
        return urls

    def generate(self, views_list):
        if r'_lang' not in views_list:
            views_list['_lang'] = {
                'view': 'vstutils.api.views.LangViewSet'
            }

        if r'user' not in views_list:
            views_list[r'user'] = {
                'view': 'vstutils.api.auth.UserViewSet'
            }

        super().generate(views_list)


class MainRouter(_AbstractRouter):
    routers = []

    def _get_custom_lists(self):
        return super()._get_custom_lists() + self.routers

    def get_api_root_view(self, *args, **kwargs):

        class API(self.APIRootView):
            if self.permission_classes:
                permission_classes = self.permission_classes
            routers = self.routers
            custom_urls = self.custom_urls
            versioning_class = None
            view_name = f'{settings.PROJECT_GUI_NAME} REST API'

            def get_view_name(self): return 'REST API'

            def get(self_inner, request, *args, **kwargs):
                # pylint: disable=no-self-argument,protected-access
                links = self._get_views_custom_list(
                    request, super(API, self_inner).get(request, *args, **kwargs)
                )
                data = {
                    'description': self_inner.view_name,
                    'current_version': links[settings.VST_API_VERSION],
                    'available_versions': {
                        name: value
                        for name, value in links.items()
                        if name in settings.API
                    }
                }
                for link, link_value in links.items():
                    if link not in data['available_versions']:
                        data[link] = link_value
                return responses.HTTP_200_OK(data)

        return API.as_view(api_root_dict=self._get_api_root_dict())

    def register_router(self, prefix, router, name=None):
        name = name or getattr(router, 'router_version_name', router.root_view_name)
        self.routers.append((prefix, router, name))

    def unregister_router(self, prefix):
        self.routers = self._unreg(prefix, self.routers)  # nocv

    def get_urls(self):
        urls = super().get_urls()
        for prefix, router, name in self.routers:
            urls.append(re_path(
                prefix,
                include(
                    (router.urls, settings.VST_PROJECT),
                    namespace=getattr(router, 'router_version_name', name)
                )
            ))
        for prefix, view, name in self.custom_urls:
            # can't be tested because this initialization takes place before
            # any test code can be run
            # pylint: disable=cell-var-from-loop
            view = getattr(view, 'as_view', lambda: view)()
            urls.append(re_path(f"^{prefix}/$", view, name=name))
        return urls

    def generate_routers(self, api):
        for version, views_list in api.items():
            router = APIRouter(perms=(permissions.IsAuthenticated,), version=version)
            router.generate(views_list)
            self.register_router(version+'/', router)

        # Register view for new bulk requests
        self.register_view(
            r'endpoint',
            import_class(settings.ENDPOINT_VIEW_CLASS),
            'endpoint'
        )



from django.db.models.base import Model
from django.db.models.manager import Manager

from .queryset import BQuerySet
from .base import ModelBaseClass


class BaseModel(Model, metaclass=ModelBaseClass):
    objects = Manager.from_queryset(BQuerySet)()

    class Meta:
        abstract = True



from copy import copy
from warnings import warn

from django.http import FileResponse
from rest_framework import status
from drf_yasg.inspectors.view import SwaggerAutoSchema
from drf_yasg.app_settings import swagger_settings
from drf_yasg.openapi import Operation

from ... import utils
from ...models.base import get_proxy_labels
from ..decorators import NestedWithAppendMixin
from .inspectors import (
    CommaMultiSelectFieldInspector,
    FkFieldInspector,
    DynamicJsonTypeFieldInspector,
    AutoCompletionFieldInspector,
    VSTFieldInspector,
    NestedFilterInspector,
    ArrayFilterQueryInspector,
    VSTReferencingSerializerInspector,
    RelatedListFieldInspector,
    RatingFieldInspector,
    NamedBinaryImageInJsonFieldInspector,
    MaskedFieldInspector,
    DecimalFieldInspector,
    CSVFileFieldInspector,
)


class VSTAutoSchema(SwaggerAutoSchema):
    field_inspectors = [
        CommaMultiSelectFieldInspector,
        FkFieldInspector,
        DynamicJsonTypeFieldInspector,
        AutoCompletionFieldInspector,
        VSTFieldInspector,
        VSTReferencingSerializerInspector,
        RelatedListFieldInspector,
        RatingFieldInspector,
        RelatedListFieldInspector,
        NamedBinaryImageInJsonFieldInspector,
        MaskedFieldInspector,
        DecimalFieldInspector,
        CSVFileFieldInspector,
    ] + swagger_settings.DEFAULT_FIELD_INSPECTORS

    filter_inspectors = [
        NestedFilterInspector,
        ArrayFilterQueryInspector,
    ] + swagger_settings.DEFAULT_FILTER_INSPECTORS

    default_status_messages: dict = {
        s[1]: ' '.join(s[2:])
        for s in map(lambda j: j.split('_'), filter(lambda x: x.startswith("HTTP_"), dir(status)))
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sch = args[0].schema
        self._sch.view = args[0]

    def _get_nested_view_class(self, nested_view, view_action_func):
        # pylint: disable=protected-access
        if not hasattr(view_action_func, '_nested_name'):
            return nested_view

        nested_action_name = '_'.join(view_action_func._nested_name.split('_')[1:])

        if nested_view is None:
            return nested_view  # nocv

        if hasattr(view_action_func, '_nested_view'):
            nested_view_class = view_action_func._nested_view
            view_action_func = getattr(nested_view_class, nested_action_name, None)
        else:  # nocv
            nested_view_class = None
            view_action_func = None

        if view_action_func is None:
            return nested_view

        return self._get_nested_view_class(nested_view_class, view_action_func)

    def __get_nested_view_obj(self, nested_view, view_action_func):
        # pylint: disable=protected-access
        # Get nested view recursively
        nested_view = utils.get_if_lazy(self._get_nested_view_class(nested_view, view_action_func))
        # Get action suffix
        replace_pattern = view_action_func._nested_subname + '_'
        replace_index = self.view.action.index(replace_pattern) + len(replace_pattern)
        action_suffix = self.view.action[replace_index:]
        # Check detail or list action
        is_detail = action_suffix.endswith('detail')
        is_list = action_suffix.endswith('list')
        # Create view object
        method = self.method.lower()
        nested_view_obj = nested_view()
        nested_view_obj.request = self.view.request
        nested_view_obj.kwargs = self.view.kwargs
        nested_view_obj.lookup_field = self.view.lookup_field
        nested_view_obj.lookup_url_kwarg = self.view.lookup_url_kwarg
        nested_view_obj.format_kwarg = None
        nested_view_obj.format_kwarg = None
        nested_view_obj._nested_wrapped_view = getattr(view_action_func, '_nested_wrapped_view', None)
        # Check operation action
        if method == 'post' and is_list:
            nested_view_obj.action = 'create'
        elif method == 'get' and is_list:
            nested_view_obj.action = 'list'
        elif method == 'get' and is_detail:
            nested_view_obj.action = 'retrieve'
        elif method == 'put' and is_detail:
            nested_view_obj.action = 'update'
        elif method == 'patch' and is_detail:
            nested_view_obj.action = 'partial_update'
        elif method == 'delete' and is_detail:
            nested_view_obj.action = 'destroy'
        else:
            nested_view_obj.action = action_suffix
            view = getattr(nested_view_obj, action_suffix, None)
            if view is not None:
                serializer_class = view.kwargs.get('serializer_class', None)
                if serializer_class:
                    nested_view_obj.serializer_class = serializer_class

        return nested_view_obj

    def get_operation_id(self, operation_keys=None):
        new_operation_keys = []
        append_new_operation_keys = new_operation_keys.append

        for key in operation_keys or []:
            append_new_operation_keys(key.replace(f'{None if not new_operation_keys else new_operation_keys[-1]}_', ''))

        return super().get_operation_id(tuple(new_operation_keys))

    def get_response_schemas(self, response_serializers):
        responses = super().get_response_schemas(response_serializers)
        for response_code, response in responses.items():
            if not response.description:
                response.description = self.default_status_messages.get(response_code, 'Action accepted.')
        return responses

    def __get_nested_view_and_subaction(self, default=None):
        sub_action = getattr(self.view, self.view.action, None)
        return getattr(sub_action, '_nested_view', default), sub_action

    def __perform_with_nested(self, func_name, *args, **kwargs):
        # pylint: disable=protected-access
        nested_view, sub_action = self.__get_nested_view_and_subaction()
        if nested_view and sub_action:
            schema = copy(self)
            try:
                schema.view = self.__get_nested_view_obj(nested_view, sub_action)
                result = getattr(schema, func_name)(*args, **kwargs)
                if result:
                    return result
            except Exception as err:  # nocv
                warn(
                    f"Error in parse '{self.view.action}'."
                    f" Using default inspection. Err: {err}"
                )
        return getattr(super(), func_name)(*args, **kwargs)

    def get_view_serializer(self, *args, **kwargs):
        return self.__perform_with_nested('get_view_serializer', *args, **kwargs)

    def get_pagination_parameters(self, *args, **kwargs):
        return self.__perform_with_nested('get_pagination_parameters', *args, **kwargs)

    def get_paginated_response(self, *args, **kwargs):
        return self.__perform_with_nested('get_paginated_response', *args, **kwargs)

    def get_filter_parameters(self, *args, **kwargs):
        return self.__perform_with_nested('get_filter_parameters', *args, **kwargs)

    def get_responses(self, *args, **kwargs):
        return self.__perform_with_nested('get_responses', *args, **kwargs)

    def get_produces(self):
        serializer = self.get_view_serializer()
        if isinstance(serializer, FileResponse):
            produces = getattr(self.view, f'produces_for_{self.view.action}', []) + ['application/octet-stream']
        else:
            produces = []
        return produces + super().get_produces()

    def get_operation(self, operation_keys=None):
        result: Operation = self.__perform_with_nested('get_operation', operation_keys)
        _nested_wrapped_view = getattr(self.view, '_nested_wrapped_view', None)
        if result['operationId'].endswith('_add') and _nested_wrapped_view:
            # pylint: disable=protected-access
            result['x-allow-append'] = issubclass(_nested_wrapped_view, NestedWithAppendMixin)
        if self.method.lower() == 'get':
            subscribe_view = self.__get_nested_view_and_subaction(self.view)[0]
            queryset = getattr(subscribe_view, 'queryset', None)
            if queryset is not None:
                # pylint: disable=protected-access
                subscribe_labels = [queryset.model._meta.label, *get_proxy_labels(queryset.model)]
                subscribe_labels += list(getattr(queryset.model, '_extra_subscribe_labels', []))
                result['x-subscribe-labels'] = subscribe_labels
            deep_nested_subview = getattr(subscribe_view, 'deep_nested_subview', None)
            if deep_nested_subview:
                result['x-deep-nested-view'] = deep_nested_subview
        else:
            if 'x-multiaction' in self.overrides:
                result['x-multiaction'] = self.overrides['x-multiaction']
            if 'x-require-confirmation' in self.overrides:
                result['x-require-confirmation'] = self.overrides['x-require-confirmation']

        return result



import typing

import orjson
from django.conf import settings


class PWAManifest:
    __slots__ = ('manifest_data',)
    default_pwa_manifest = {
        "name": settings.PROJECT_GUI_NAME.capitalize(),
        "short_name": settings.VST_PROJECT.capitalize(),
        "theme_color": "rgb(236,240,245)",
        "background_color": "rgb(236,240,245)",
        "display": "standalone",
        "scope": ".",
        "start_url": "./",
        "icons": []
    }
    default_pwa_manifest.update(getattr(settings, 'PWA_MANIFEST', {}))

    def __init__(self, override_defaults: typing.Dict = None):
        self.manifest_data = self.default_pwa_manifest.copy()
        self.manifest_data.update(override_defaults or {})

    @property
    def icons(self):
        return tuple(self.manifest_data.get("icons", []))

    @property
    def data(self) -> typing.Dict:
        return self.manifest_data

    @property
    def json(self) -> typing.Text:
        return orjson.dumps(self.data, option=orjson.OPT_INDENT_2).decode('utf-8')



# pylint: disable=no-member,no-classmethod-decorator,protected-access
from django.db import models
from django.db.models.expressions import RawSQL
from django.utils.functional import cached_property
from django.conf import settings

from ..utils import Paginator, raise_context_decorator_with_default, is_member_descriptor


class BQuerySet(models.QuerySet):
    """
    Represent a lazy database lookup for a set of objects.
    Allows to override default iterable class by `custom_iterable_class` attribute
    (class with `__iter__` method which returns generator of model objects) and
    default query class by `custom_query_class` attribute
    (class inherited from :class:`django.db.models.sql.query.Query`).
    """

    __slots__ = ('__iterable_class__',)
    use_for_related_fields = True
    custom_query_class = None

    def __init__(self, model=None, query=None, using=None, hints=None):
        if query is None and self.custom_query_class is not None:
            query = self.custom_query_class(self)  # pylint: disable=not-callable
        super().__init__(model=model, query=query, using=using, hints=hints)

    @property
    def _iterable_class(self):
        if hasattr(self, '__iterable_class__') and not is_member_descriptor(self.__iterable_class__):
            return self.__iterable_class__
        if hasattr(self, 'custom_iterable_class'):  # pragma: no branch
            self.__iterable_class__ = self.custom_iterable_class
        return self._iterable_class

    @_iterable_class.setter
    def _iterable_class(self, value):
        if not hasattr(self, 'custom_iterable_class'):
            self.__iterable_class__ = value

    @_iterable_class.deleter
    def _iterable_class(self):  # nocv
        del self.__iterable_class__

    def paged(self, *args, **kwargs):
        """
        Returns paginated data with custom Paginator-class.
        Uses `PAGE_LIMIT` from global settings by default.
        """
        return self.get_paginator(*args, **kwargs).items()

    def get_paginator(self, *args, **kwargs):
        """
        Returns initialized object of :class:`vstutils.utils.Paginator` over
        current instance's QuerySet. All args and kwargs go to to Paginator's constructor.
        """
        return Paginator(self.filter(), *args, **kwargs)

    def has_field_filter_in_query(self, field_name):
        return any(filter(
            raise_context_decorator_with_default(default=False)(
                lambda x: getattr(x.lhs.field, 'attname', None) == field_name
            ),
            self.query.where.children
        ))

    @cached_property
    def has_hidden_filter(self):
        return self.has_field_filter_in_query('hidden')

    def cleared(self):
        """
        Filter queryset for models with attribute 'hidden' and
        exclude all hidden objects.
        """
        if hasattr(self.model, "hidden") and not self.has_hidden_filter:
            return self.filter(hidden=False)
        return self

    def _get_deep_nested_qs_with_cte(self, with_current=False, deep_children=True):
        model_to_sql = self.model
        origin_model_pk = field_to_get = self.model._meta.pk.name
        deep_field_name = self.model.deep_parent_field

        field = getattr(self.model, deep_field_name)
        m2m_model = getattr(field, 'through', None)
        if m2m_model:
            field_to_get, deep_field_name = getattr(
                self.model._meta.get_field(deep_field_name),
                'through_fields',
                [f'from_{self.model.__name__.lower()}_id', f'to_{self.model.__name__.lower()}_id']
            )
            model_to_sql = m2m_model

        sql_table = model_to_sql._meta.db_table
        sql_column_to_get = getattr(model_to_sql, field_to_get).field.column
        sql_deep_column = getattr(model_to_sql, deep_field_name).field.column

        if not deep_children:
            # then deep_parents
            sql_column_to_get, sql_deep_column = sql_deep_column, sql_column_to_get

        with_current_sql = f'''
            UNION
            {self.values(origin_model_pk).order_by().query}
        '''

        sql = f'''
            WITH RECURSIVE nested as (
                    SELECT {sql_table}.{sql_column_to_get}, {sql_table}.{sql_deep_column}
                    FROM {sql_table}
                    WHERE {sql_table}.{sql_deep_column}
                        IN ({str(self.values(origin_model_pk).order_by().query)})
                UNION
                    SELECT {sql_table}.{sql_column_to_get}, {sql_table}.{sql_deep_column}
                    FROM {sql_table}
                        JOIN nested
                            ON {sql_table}.{sql_deep_column} = nested.{sql_column_to_get}
            )
            SELECT {sql_column_to_get} from nested
        '''
        if with_current:
            sql += with_current_sql
        return self.model.objects.filter(id__in=RawSQL(sql, []))  # nosec

    def _deep_nested_ids_without_cte(self, accumulated=None, deep_children=True):
        deep_parent_field = self.model.deep_parent_field
        related_name = self.model._meta.get_field(deep_parent_field).related_query_name()
        if deep_children:
            related_name = deep_parent_field

        # accumulate retrieved id's
        accumulated = accumulated if accumulated else self.none()
        # get all id's from model except already accumulated
        list_id = self.exclude(id__in=accumulated).values_list("id", flat=True)
        accumulated = (accumulated | list_id)
        kw = {related_name + "__id__in": list_id}
        subs = self.model.objects.using(self.db).filter(**kw)
        subs_id = subs.values_list("id", flat=True)
        if subs_id:
            accumulated = (accumulated | subs._deep_nested_ids_without_cte(accumulated, deep_children))
        return accumulated

    def _get_deep_qs(self, with_current=False, deep_children=True):
        if self.db in settings.DATABASES_WITHOUT_CTE_SUPPORT:
            qs = self.model.objects.filter(id__in=self._deep_nested_ids_without_cte(deep_children=deep_children))
            if not with_current:
                qs = qs.exclude(id__in=self.values('id'))
            return qs
        return self._get_deep_nested_qs_with_cte(with_current=with_current, deep_children=deep_children)

    def get_parents(self, with_current=False):
        return self._get_deep_qs(with_current=with_current, deep_children=False)

    def get_children(self, with_current=False):
        return self._get_deep_qs(with_current=with_current, deep_children=True)



# flake8: noqa: E501
TRANSLATION = {
    'new': 'Новый',
    'Create': 'Создать',
    'add': 'добавить',
    'Edit': 'Редактировать',
    'edit': 'Редактировать',
    'Save': 'Сохранить',
    'Remove': 'Удалить',
    'Reload': 'Перезагрузить',
    'send': 'отправить',
    'select': 'выбрать',
    'filters': 'фильтры',
    'apply': 'применить',
    'cancel': 'отменить',
    'field': 'поле',
    'actions': 'действия',
    'sublinks': 'вложенные ссылки',
    'name': 'наименование',
    'type': 'тип',
    'owner': 'владелец',
    'notes': 'заметки',
    'copy': 'копировать',
    'error': 'ошибка',
    'home': 'главная',
    'login': 'войти',
    'logout': 'выйти',
    'Profile': 'Профиль',
    'settings': 'настройки',
    'Settings': 'Настройки',
    'version': 'версий | версия | версии',
    'cache': 'кэш',
    'documentation': 'документация',
    'request': 'запрос',
    'repository': 'репозиторий',
    'app info': 'инфо',
    'status': 'статус',
    'List is empty': 'Список пуст',
    'child instances': 'вложенные объекты',
    'enter value': 'введите значение',
    'more info': 'подробнее',
    'search by': 'поиск по полю',
    'Search': 'Поиск',
    'Selected': 'Выбрано',
    'Nothing found!': 'Ничего не найдено!',
    'Main settings': 'Основные настройки',
    'Custom settings': 'Другие настройки',
    'Remove from list': 'Убрать из списка',
    'Purge': 'Удалить',
    'Select an action for this instances': 'Выберите действие для объектов',
    'Username not': 'Имя пользователя не',

    # fields
    'Id': 'Идентификатор',
    'ID': 'Идентификатор',

    # x-menu generic
    'system': 'Система',
    'users': 'Пользователи',
    'Is active': 'Активен',
    'Is staff': 'Модератор',

    # multiactions button
    'actions on': 'действия над',
    'on item': 'элементов | элементом | элементами | элементами',

    'key': 'ключ',
    'value': 'значение',
    'new owner': 'новый владелец',
    'detail': 'детали',
    'description': 'описание',
    'username': 'никнейм',
    'is active': 'активен',
    'user': 'пользователь',
    'yes': 'да',
    'no': 'нет',
    'is staff': 'админ',
    'first name': 'имя',
    'last name': 'фамилия',
    'password': 'пароль',
    'Password': 'Пароль',
    'Change password': 'Сменить пароль',
    'Change_password': 'сменить пароль',
    'generate password': 'сформировать пароль',
    'repeat password': 'повторить пароль',
    'Repeat password': 'Повторить пароль',
    'Old password': 'Старый пароль',
    'New password': 'Новый пароль',
    'Confirm new password': 'Подтвердить новый пароль',
    'kind': 'вид',
    'mode': 'режим',
    'options': 'опции',
    'information': 'информация',
    'date': 'дата',

    'file n selected': 'нет файлов | 1 файл выбран | {n} файла выбрано | {n} файлов выбрано',
    'image n selected': 'нет изображений | 1 изображение выбрано | {n} изображения выбрано | {n} изображений выбрано',

    # filters
    'Ordering': 'Порядок сортировки',
    'is_active': 'активен',
    'Username': 'Ник',
    'First name': 'Имя',
    'Last name': 'Фамилия',

    'Page matching current url was not found': 'Cтраница, соответствующая данному url, не была найдена',
    'Unknown error': 'Неизвестная ошибка',
    'Error': 'Ошибка',

    'We\'ve emailed you instructions for setting your password, if an account exists with the email you entered. You should receive them shortly.': 'Мы отправили вам письмо с инструкцией для сброса пароля.',
    'If you don\'t receive an email, please make sure you\'ve entered the address you registered with, and check your spam folder.': 'Если вы не получили письмо, пожалуйста, убедитесь, что введен верный адрес, и проверьте папку со спамом.',
    'The password reset link was invalid, possibly because it has already been used. Please request a new Password reset.': 'Неверная ссылка для восстановления пароля, возможно она уже была использована. Пожалуйста, попробуйте заново.',
    'Please enter your new password twice so we can verify you typed it in correctly.': 'Пожалуйста, введите новый пароль.',
    'change my password': 'изменить пароль',
    'Your password has been set.  You may go ahead and log in now.': 'Ваш пароль был успешно установлен. Теперь вы можете войти.',
    'Please enter a correct %(username)s and password. Note that both fields may be case-sensitive.': 'Пожалуйста, введите корректные %(username)s и пароль. Внимание, поля могуь быть регистрозависимыми.',
    'or': 'или',
    'two-factor authentication code': 'код двухфакторной авторизации',
    'verify code': 'подтвердить код',
    'Enter the code from the two-factor app on your mobile device. If you\'ve lost your device, you may enter one of your recovery codes.': 'Введите код из двухфакторного приложения на мобильном устройстве. Если вы потеряли свое устройство, вы можете ввести один из кодов восстановления.',
    'Scan the image with the two-factor authentication app on your phone. If you can’t use a barcode, enter this text code instead.': 'Отсканируйте изображение с помощью приложения для двухфакторной аутентификации на вашем телефоне. Если вы не можете использовать штрих-код, введите вместо него текстовый код.',
    'Enter the six-digit code from the application': 'Введите шестисимвольный код из приложения',
    'code': 'код',
    'recovery codes': 'коды восстановления',
    'Recovery codes are used to access your account in the event you cannot receive two-factor authentication codes.': 'Коды восстановления используются для доступа к вашей учетной записи в случае, если вы не можете получить коды двухфакторной аутентификации.',
    'Disabling Two-Factor Authentication (2FA) will decrease your account\'s security': 'Отключение двухфакторной авторизации снизит безопасность вашей учетной записи.',
    'Enable': 'Включить',
    'Disable': 'Выключить',
    'Two factor authentication': 'Двухфакторная авторизация',
    'No {0} matches the given query.': "Ни один {0} не соответствует данному запросу.",
    'Password is not correct.': 'Неверный пароль.',
    'New passwords values are not equal.': 'Значения нового пароля не эквивалентны.',
    'This password is entirely numeric.': 'Этот пароль полностью числовой.',
    'Passwords do not match.': 'Пароли не совпадают.',
    "The two password fields didn't match.": 'Два поля пароля не совпадают.',
    'Dark theme': 'Темная тема',
    'Language': 'Язык',
    'height': 'высота',
    'width': 'ширина',
    'More info': 'Подробнее',
    'Dark mode': 'Темная тема',
    'Execute': 'Выполнить',
    'Domain should be setuped or username should be with @domain.name': "Домен должен быть настроен или логин должен содержать в конце @имя.домена",
    'Invalid name': 'Неправильное имя',
    'Should be [full.domain.name]': 'Должно быть в формате [полное.доменное.имя].',
    'Invalid registration email send.': 'Отправлена неверная регистрационная почта.',
    'To continue, need to accept the terms agreement.': 'Чтобы продолжить, необходимо принять условия соглашения.',

    # Проверить
    'Changes in settings are successfully saved. Please refresh the page.': 'Изменения в настройках успешно сохранены. Пожалуйста, обновите страницу.',
    'now': 'сейчас',
    'later': 'позже',
    'I accept the ': 'Я принимаю ',
    'Zoom': 'Зум',
    'Scale': 'Масштаб',
    'This file format is not supported': 'Данный формат файла не поддерживается',
    'Oops! Something went wrong.': 'Упс! Что-то пошло не так',
    '[Object not found]': '[Объект не найден]',
    'Execute action on {0}': 'Выполнить действие для {0}',
    '{n} instance': '0 объектов | {n} объекта | {n} объектов',
    'Open image': 'Открыть изображение',
    'Confirm action': 'Подтвердить действие',
    'Confirm': 'Подтвердить',
    'Cancel': 'Отменить',

    # field's validation
    # empty
    'Field "<b>{0}</b>" is empty.': 'Поле "<b>{0}</b>" пустое.',
    # required
    'Field "<b>{0}</b>" is required.': 'Поле "<b>{0}</b>" обязательно для заполнения.',
    # minLength
    'Field "<b>{0}</b>" is too short.<br> Field length should not be shorter, then {1}.': 'Длина поля "<b>{0}</b>" слишком мала.<br>Длина поля должна быть не меньше, чем {1}.',
    # maxLength
    'Field "<b>{0}</b>" is too long. <br> Field length should not be longer, then {1}.': 'Длина поля "<b>{0}</b>" слишком велика.<br>Длина поля должна быть не больше, чем {1}.',
    # min
    'Field "<b>{0}</b>" is too small.<br> Field should not be smaller, then {1}.': 'Значение поля "<b>{0}</b>" слишком мало.<br> Значение поля должно быть не меньше, чем {1}.',
    # max
    'Field "<b>{0}</b>" is too big.<br> Field should not be bigger, then {1}.': 'Значение поля "<b>{0}</b>" слишком велико.<br> Значение поля должно быть не больше, чем {1}.',
    # invalid
    '<b>{0} </b> value is not valid for <b>{1}</b> field.': 'Значение <b>{0}</b> не допустимо для поля <b>{1}</b>.',
    # email
    '<b>"{0}"</b> field should be written in <b>"example@mail.com"</b> format.': 'Поле <b>"{0}"</b> должно быть заполнено в формате <b>"example@mail.com"</b>.',

    # validators
    'imageValidationResolutionError': 'Должно быть не меньше {min} и не больше {max} px.',

    # instance operation success
    # add
    'Child "<b>{0}</b>" instance was successfully added to parent list.': 'Дочерний объект "<b>{0}</b>" был успешно добавлен в родительский список.',
    # create
    'New "<b>{0}</b>" instance was successfully created.': 'Новый объект "<b>{0}</b>" был успешно создан.',
    # remove
    '"<b>{0}</b>" {1} was successfully removed.': '"<b>{0}</b>" {1} был(а) успешно удален.',
    # save
    'Changes in "<b>{0}</b>" {1} were successfully saved.': 'Изменения в объекте {1} "<b>{0}</b>" были успешно сохранены.',
    # execute
    'Action "<b>{0}</b>" was successfully executed on "<b>{1}</b>" instance.': 'Действие "<b>{0}</b>" было успешно запущено на объекте "<b>{1}</b>".',
    # instance operation error
    # add:
    'An error occurred during adding of child "<b>{0}</b>" instance to parent list.<br> Error details: {1}': 'Во время добавления дочернего объекта "<b>{0}</b>" к родительскому списку произошла ошибка.<br> Подробнее: {1}',
    # create:
    'An error occurred during creation.<br>Error details:<br>{0}': 'Во время создания нового объекта произошла ошибка.<br>Подробнее:<br>{0}',
    # remove:
    'An error occurred during removal process of "<b>{0}</b>" {1}.<br> Error details: {2}': 'Во время удаления {1} "<b>{0}</b>" произошла ошибка.<br> Подробнее: {2}',
    # save:
    'An error occurred during process.<br>Error details:<br>{0}': 'Во время сохранения произошла ошибка.<br>Подробнее:<br>{0}',
    # execute:
    'An error occurred during <b>{0}</b>. Error details:<br>{1}': 'Во время запуска действия <b>{0}</b> произошла ошибка. Подробнее:<br>{1}',
    'Link': 'Ссылка',
    # csv
    'Actions': 'Действия',
    'Index': 'Индекс',
    'Download with headers': 'Скачать с заголовками столбцов',
}

SERVER_TRANSLATION = {
    # field's validation
    "value is not a valid list": "значение не является допустимым списком",
    "value is not a valid JSON": "значение не является допустимым JSON",
    "key {missing_key} is missing": "ключ {missing_key} отсутствует",
    "invalid key {invalid_key}": "неверный ключ {invalid_key}",

    # field help_text
    'A confirmation will be sent to your e-mail': 'На эту почту будет отправлено письмо для подтверждения регистрации',
    'Required. Inform a valid email address': 'Напишите вашу электронную почту',
    # validators
    'unsupported image file format, expected ({0}), got {1}': 'Неподдерживаемый формат изображения. Поддерживается ({0}), получен {1}',
    'For some reason, this image file cannot be opened': 'Не удалось открыть изображение',
    'Invalid image {0}. Expected from {1} to {2}, got {3}': 'Некорректная {0} изображения. Ожидалось от {1} до {2}, получено {3}',
    'This field must contain only digits but input: ': 'Это поле должно содержать только цифры, но введено: ',
    'Unsupported image file format': 'Не поддерживаемый: формат файла',
    'is not in listed supported types': 'не представлен в списке поддерживаемых типов',
    'Invalid image size orientations': 'Неправильные размеры изображения',
    'Current image size': 'Текущий размер изображения',

    # agreements
    'personal data processing policy': 'Политика обработки персональных данных',
    'I agree with ': 'Я согласен с ',
    'the personal data processing policy': 'политикой обработки персональных данных',
    'To continue, need to agree to the personal data processing policy.': 'Чтобы продолжить, необходимо согласиться '
                                                                          'с политикой обработки персональных данных.',
    'terms of agreement': 'лицензионное соглашение',

    # auth/reg
    'sign in to start your session': 'Войдите, чтобы начать сеанс',
    'confirm your email before logging in': 'подтвердите свою электронную почту перед входом',
    'Confirmation link is invalid or expired': 'Ссылка для подтверждения неверна или устарела',
    'sign in': 'войти',
    'sign up': 'зарегистрируйтесь',
    'forgot password': 'забыли пароль',
    'confirm password': 'подтвердите пароль',
    'register': 'регистрация',
    'fields with * is required': 'поля отмеченые * обязательны для заполнения',
    'Forgot your password? Enter your email address below, and an email with instructions for setting a new one will be sent.': 'Забыли пароль? Введите свой адрес, и мы отправим вам инструкцию для восстановления.',
    'reset my Password': 'сбросить пароль',
    'Invalid authentication code': 'Неверный код авторизации',
    'Secret string must be provided': 'Секретная строка должна быть заполнена',

    # filters help text
    'A unique integer value (or comma separated list) identifying this instance.': 'Уникальное числовое значение (или их последовательность разделенная запятой) идентифицирующая данный объект.',
    'A name string value (or comma separated list) of instance.': 'Имя объекта - строковое значение (или их последовательность разделенная запятой).',
    'Which field to use when ordering the results.': 'Поле, по которому производить сортировку результатов.',
    "Search by {}'s primary key or {}": 'Искать по ключу {} или {}',
    '_name': 'наименование',
    'Boolean value meaning status of user.': 'Значение, обозначающее статус пользователя.',
    'Users first name.': 'Имя пользователя',
    'Users last name.': 'Фамилия пользователя',
    'Users e-mail value.': 'Email пользователя',
}



import typing as _t
import logging
import uuid

from django.db.models import signals
from django.conf import settings
from django.contrib.auth import get_user_model
from cent import Client as CentrifugoClient  # type: ignore

from .base import get_proxy_labels
from ..utils import raise_context_decorator_with_default, raise_context
from .model import BaseModel


logger = logging.getLogger('vstutils')


class Notificator:
    client_class = CentrifugoClient
    default_channel = "subscriptions_update"

    queue: _t.List[_t.Tuple[_t.Sequence[_t.Text], _t.Any]]
    cent_client: CentrifugoClient
    channel: _t.Text

    def __init__(self, queue=None, client=None, channel=None):
        self.queue = queue or []
        self.cent_client = client or self.get_client()
        self.channel = channel or self.default_channel
        self._signals: _t.List[signals.ModelSignal] = []
        if self.cent_client is not None:
            self.connect_signal(signals.post_save)
            self.connect_signal(signals.post_delete)
        logger.debug(f'Notificator for channel {self.channel} initialized.')

    def connect_signal(self, signal: signals.ModelSignal):
        if signal not in self._signals:
            signal.connect(self.signal_handler)
            self._signals.append(signal)

    def disconnect_signal(self, signal: signals.ModelSignal):
        if signal in self._signals:
            signal.disconnect(self.signal_handler)

    def signal_handler(self, instance, *args, **kwargs):
        if isinstance(instance, (BaseModel, get_user_model())) and getattr(instance, '_notify_update', True):
            self.create_notification_from_instance(instance)

    @raise_context_decorator_with_default(verbose=False)
    def get_client(self):
        centrifugo_client_kwargs = {**settings.CENTRIFUGO_CLIENT_KWARGS}
        centrifugo_client_kwargs.pop('token_hmac_secret_key', None)
        logger.debug(f"Getting Centrifugo client with kwargs: {centrifugo_client_kwargs}")
        return self.client_class(**centrifugo_client_kwargs)

    def create_notification_from_instance(self, instance):  # pylint: disable=invalid-name
        model = instance.__class__
        self.queue.append(
            ((model._meta.label, *get_proxy_labels(model)), instance.pk)
        )

    def create_notification(self, label, pk):
        if isinstance(label, str):
            label = (label,)
        self.queue.append(
            (label, pk)
        )

    @raise_context_decorator_with_default()
    def send(self):
        self.queue, objects = [], set(self.queue)

        for labels, pk in objects:
            if isinstance(pk, uuid.UUID):
                pk = str(pk)
            with raise_context():
                self.cent_client.add("publish", self.cent_client.get_publish_params(
                    self.channel,
                    {
                        "subscribe-label": labels,
                        "pk": pk
                    }
                ))
        if objects:
            logger.debug(f'Send notifications about {len(objects)} updates.')
            return self.cent_client.send()

    def disconnect_all(self):
        for signal in self._signals:
            self.disconnect_signal(signal)

    def __del__(self):
        self.disconnect_all()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.send()
        self.disconnect_all()



# pylint: disable=django-not-available,import-outside-toplevel
import os
import sys

_default_settings = {
    # vstutils settings for generate settings
    "VST_PROJECT": "vstutils",
    'VST_DEV_SETTINGS': 'dev_settings.ini',
    # django settings module
    "DJANGO_SETTINGS_MODULE": 'vstutils.settings',
    # Default urlconf
    "VST_ROOT_URLCONF": "vstutils.urls",
}


def prepare_environment(default_settings=None, **kwargs):
    # pylint: disable=unused-argument
    """
    Prepare ENV for web-application
    :param default_settings: minimal needed settings for run app
    :type default_settings: dict
    :param kwargs: other overrided settings
    :rtype: None
    """
    default_settings = _default_settings if default_settings is None else default_settings
    for key, value in default_settings.items():
        os.environ.setdefault(key, value)
    os.environ.update(kwargs)


def cmd_execution(*args, **kwargs):
    # pylint: disable=unused-variable
    """
    Main function to executes from cmd. Emulates django-admin.py execution.
    :param kwargs: overrided env-settings
    :rtype: None
    """
    from django.core.management import execute_from_command_line
    prepare_environment(**kwargs)
    argv = list(args or sys.argv)
    argv[0] = os.getenv("VST_CTL_SCRIPT", sys.argv[0])
    execute_from_command_line(argv or sys.argv)


def get_celery_app(name=None, **kwargs):  # nocv
    # pylint: disable=import-error
    """
    Function to return celery-app. Works only if celery installed.
    :param name: Application name
    :param kwargs: overrided env-settings
    :return: Celery-app object
    """
    from celery import Celery
    prepare_environment(**kwargs)
    name = name or os.getenv("VST_PROJECT")
    celery_app = Celery(name)
    celery_app.config_from_object('django.conf:settings', namespace='CELERY')
    celery_app.autodiscover_tasks()
    return celery_app



# Compilation block
########################################################################################
import re
import os
import sys
import fnmatch
import codecs
import gzip
import glob
import shutil

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

from setuptools import find_packages, setup, Command
from setuptools.extension import Extension
from setuptools.command.sdist import sdist as _sdist
from setuptools.command.build_py import build_py as build_py_orig
from setuptools.command.install_lib import install_lib as _install_lib
try:
    from Cython.Build import cythonize, build_ext as _build_ext
except ImportError:
    has_cython = False
else:
    has_cython = True

try:
    from sphinx.setup_command import BuildDoc
    import sphinx  # noqa: F401
    has_sphinx = True
except ImportError:
    has_sphinx = False


ignored_keys = ['-h', '--help', '--version']
is_help = any([a for a in ignored_keys if a in sys.argv])
is_develop = 'develop' in sys.argv
is_build = (any([a for a in ['compile', 'bdist_wheel', 'bdist'] if a in sys.argv]) or is_develop) and not is_help


def get_discription(file_path='README.rst', folder=os.getcwd()):
    with codecs.open("{}/{}".format(folder, file_path), 'r', encoding='utf-8') as readme:
        return readme.read()


def load_requirements(file_name, folder=os.getcwd()):
    with codecs.open(os.path.join(folder, file_name), 'r', encoding='utf-8')as req_file:
        return req_file.read().strip().split('\n')


def get_file_ext(ext):
    file_types = [".py", ".pyx", ".c", '.cpp'] if has_cython else [".c", '.cpp', ".py"]
    for ftype in file_types:
        fname = ext.replace(".", "/") + ftype
        if os.path.exists(fname):
            return fname
    return None


def listfiles(folder):
    if not isinstance(folder, (list, tuple)):
        folder = [folder]
    folder = filter(lambda p: os.path.isdir(p), folder)
    for one_folder in folder:
        for root, folders, files in os.walk(one_folder):
            for filename in folders + files:
                yield os.path.join(root, filename)


def clear_old_extentions(extensions_list, packages):
    for filename in listfiles(packages):
        _filename, _f_ext = os.path.splitext(filename)
        if os.path.isdir(_filename) or _f_ext not in ['.c', '.cpp']:
            continue
        has_py = (
            os.path.exists('{}.py'.format(_filename)) or
            os.path.exists('{}.pyx'.format(_filename))
        )

        if has_py and filename.replace('/', '.').replace(_f_ext, '') in extensions_list:
            print('Removing old extention [{}].'.format(filename))
            os.remove(filename)


def make_extention(module_name, files, extra_compile_args, main_include_dir=os.path.join(os.getcwd(), 'include')):
    include_dirs = list(filter(
        lambda f: bool(f) and os.path.exists(f) and os.path.isdir(f),
        [os.path.join(module_name.split('.')[0], 'include'), main_include_dir]
    ))

    return Extension(
        module_name, files,
        extra_compile_args=extra_compile_args,
        include_dirs=include_dirs
    )


def make_extensions(extensions_list, packages):
    if not isinstance(extensions_list, list):
        raise Exception("Extension list should be `list`.")

    if not is_help:
        clear_old_extentions(extensions_list, packages)

    extensions_dict = {}
    for ext in extensions_list:
        files = []
        module_name = ext
        if isinstance(ext, (list, tuple)):
            module_name = ext[0]
            for file_module in ext[1]:
                file_name = get_file_ext(file_module)
                files += [file_name] if file_name else []
        else:
            file_name = get_file_ext(ext)
            files += [file_name] if file_name else []
        if files:
            extensions_dict[module_name] = files

    extra_compile_args = [
        '-g0', '-ggdb1',
        "-fno-strict-aliasing",
        "-fno-var-tracking-assignments",
        "-pipe", "-std=c99", '-Werror=sign-compare'
    ]
    ext_modules = list(
        make_extention(m, f, extra_compile_args)
        for m, f in extensions_dict.items()
    )
    ext_count = len(ext_modules)
    nthreads = ext_count if ext_count < 10 else 10

    language_level = 3
    if is_help:
        pass
    elif has_cython and ('compile' in sys.argv or 'bdist_wheel' in sys.argv or 'build_ext' in sys.argv):
        cy_kwargs = dict(
            nthreads=nthreads,
            force=True,
            language_level=language_level,
            compiler_directives=dict(
                linetrace='CYTHON_TRACE_NOGIL' in sys.argv,
                profile=True,
                c_string_type='str',
                c_string_encoding='utf8'
            ),
        )
        return cythonize(ext_modules, **cy_kwargs), extensions_dict
    return ext_modules, extensions_dict


def minify_js_file(js_file, jsmin_func):
    return jsmin_func(js_file, quote_chars="'\"`")


def minify_css_file(css_file, cssmin_func):
    return cssmin_func(css_file)


def minify_static_files(base_dir, files, exclude=None):
    exclude = exclude or []
    patterns = dict()
    try:
        from jsmin import jsmin as jsmin_func
        patterns['*.js'] = (minify_js_file, jsmin_func)
    except:
        pass
    try:
        from csscompressor import compress as csscompressor_func
        patterns['*.css'] = (minify_css_file, csscompressor_func)
    except:
        pass

    regex_exclude = [re.compile(r, re.MULTILINE) for r in exclude]

    for fnext, funcs in patterns.items():
        for fext_file in filter(lambda f: fnmatch.fnmatch(f, fnext), files):
            if fnmatch.fnmatch(fext_file, '*.min.*'):
                continue
            fext_file = os.path.join(base_dir, fext_file)
            if os.path.exists(fext_file):
                if not any(filter(lambda fp: bool(fp.search(fext_file)), regex_exclude)):
                    func, subfunc = funcs
                    with codecs.open(fext_file, 'r', encoding='utf-8') as static_file_fd:
                        minified = func(static_file_fd.read(), subfunc)
                    with codecs.open(fext_file, 'w', encoding='utf-8') as static_file_fd:
                        static_file_fd.write(minified)
                    print('Minfied file {fext_file}.'.format(fext_file=fext_file))
                with open(fext_file, 'rb') as f_in:
                    with gzip.open("{}.gz".format(fext_file), 'wb') as f_out:
                        shutil.copyfileobj(f_in, f_out)
                print('Compressed file {fext_file}.'.format(fext_file=fext_file))


def compile_py_func(fullname, compile_file_func):
    if compile_file_func(fullname, ddir=os.path.dirname(fullname), legacy=True, optimize=0):
        os.remove(fullname)


def compile_python_sources(base_dir, files, exclude=None):
    exclude = exclude or []
    patterns = dict()
    try:
        from compileall import compile_file
        patterns['*.py'] = (compile_py_func, compile_file)
    except:
        pass

    regex_exclude = [re.compile(r, re.MULTILINE) for r in exclude]

    for fnext, funcs in patterns.items():
        for fext_file in filter(lambda f: fnmatch.fnmatch(f, fnext), files):
            fext_file = os.path.join(base_dir, fext_file)
            if os.path.exists(fext_file):
                if not any(filter(lambda fp: bool(fp.search(fext_file)), regex_exclude)):
                    func, subfunc = funcs
                    funcs[0](fext_file, funcs[1])
                    print('Compiled {fext_file}.'.format(fext_file=fext_file))


class _Compile(_sdist):
    extensions_dict = dict()
    static_exclude = []

    def __filter_files(self, files):
        for _files in self.extensions_dict.values():
            for file in _files:
                if file in files:
                    files.remove(file)
        return files

    def make_release_tree(self, base_dir, files):
        if has_cython:
            files = self.__filter_files(files)
        _sdist.make_release_tree(self, base_dir, files)
        minify_static_files(base_dir, files, self.static_exclude)

    def run(self):
        return _sdist.run(self)


class GithubRelease(Command):
    '''
    Make release on github via githubrelease
    '''
    description = 'Make release on github via githubrelease'

    user_options = [
        ('body=', 'b', 'Body message.'),
        ('assets=', 'a', 'Release assets patterns.'),
        ('repo=', 'r', 'Repository for release.'),
        ('release=', 'R', 'Release version.'),
        ('dry-run=', 'd', 'Dry run.'),
        ('publish=', 'p', 'Publish release or just create draft.'),
    ]

    def initialize_options(self):
        self.body = None or os.getenv('CI_COMMIT_DESCRIPTION', None)
        self.assets = None
        self.repo = None
        self.dry_run = False
        self.publish = False
        self.release = None or self.distribution.metadata.version

    def finalize_options(self):
        if self.repo is None:
            raise Exception("Parameter --repo is missing")
        if self.release is None:
            raise Exception("Parameter --release is missing")
        self._gh_args = (self.repo, self.release)
        self._gh_kwargs = dict(
            publish=self.publish, name=self.release, dry_run=self.dry_run
        )
        if self.assets:
            assets = self.assets.format(release=self.release)
            assets = list(filter(bool, assets.split('\n')))
            self._gh_kwargs['asset_pattern'] = assets
        if self.body:
            self._gh_kwargs['body'] = self.body

    def run(self):
        from github_release import gh_release_create
        gh_release_create(*self._gh_args, **self._gh_kwargs)


class build_py(build_py_orig):
    exclude = []
    compile_extentions_types = ['.py', '.pyx']
    wheel_extentions_types = ['.c', '.cpp'] + compile_extentions_types

    def _filter_modules(self, module_tuple):
        pkg, mod, file = module_tuple
        try:
            file_name, file_ext = os.path.splitext(file)
            module_name = file_name.replace('/', '.')
        except:
            return True
        if 'bdist_wheel' in sys.argv:
            exclude_list = self.wheel_extentions_types
        elif 'compile' in sys.argv:
            exclude_list = self.compile_extentions_types
        else:
            return True
        if module_name in self.exclude and file_ext in exclude_list:
            return False
        return True

    def find_package_modules(self, package, package_dir):
        modules = build_py_orig.find_package_modules(self, package, package_dir)
        return list(filter(self._filter_modules, modules))


class install_lib(_install_lib):
    exclude = []
    static_exclude = []
    compile_exclude = []

    def _filter_files_with_ext(self, filename):
        _filename, _fext = os.path.splitext(filename)
        if _fext in build_py.wheel_extentions_types:
            return True
        return False

    def install(self):
        result = _install_lib.install(self)
        files = list(listfiles(self.install_dir))
        so_extentions = list(filter(lambda f: fnmatch.fnmatch(f, '*.so'), files))
        for source in filter(self._filter_files_with_ext, files):
            _source_name, _source_ext = os.path.splitext(source)
            if any(filter(lambda f: fnmatch.fnmatch(f, _source_name+"*.so"), so_extentions)):
                print('Removing extention sources [{}].'.format(source))
                os.remove(source)
        minify_static_files('', files, self.static_exclude)
        if os.getenv('BUILD_COMPILE', '') == 'true':
            compile_python_sources('', files, self.compile_exclude)
        return result


def get_compile_command(extensions_dict=None):
    extensions_dict = extensions_dict or dict()
    compile_class = _Compile
    compile_class.extensions_dict = extensions_dict
    return compile_class


def make_setup(**opts):
    if 'packages' not in opts:
        opts['packages'] = find_packages()
    ext_modules_list = opts.pop('ext_modules_list', [])
    ext_mod, ext_mod_dict = make_extensions(ext_modules_list, opts['packages'])
    opts['ext_modules'] = opts.get('ext_modules', []) + ext_mod
    cmdclass = opts.get('cmdclass', dict())
    static_exclude = opts.pop('static_exclude_min', [])
    if 'compile' not in cmdclass:
        compile_class = get_compile_command(ext_mod_dict)
        compile_class.static_exclude = static_exclude
        cmdclass.update({"compile": get_compile_command(ext_mod_dict)})
    if has_cython:
        build_py.exclude = ext_modules_list
        install_lib.static_exclude = static_exclude
        install_lib.compile_exclude = opts.pop('compile_modules_exclude', [])
        cmdclass.update({
            'build_ext': _build_ext,
            'build_py': build_py,
            'install_lib': install_lib
        })
    if has_sphinx and 'build_sphinx' not in cmdclass:
        cmdclass['build_sphinx'] = BuildDoc
    cmdclass['githubrelease'] = GithubRelease
    opts['cmdclass'] = cmdclass

    webpack_path = os.path.join(os.getcwd(), 'webpack.config.js')
    if os.path.exists(webpack_path) and is_build and os.environ.get('DONT_YARN', "") != 'true':
        yarn_build_command = 'devBuild' if is_develop else 'build'
        try:
            os.system('yarn install --pure-lockfile')
            os.system('yarn ' + yarn_build_command)
        except Extension as err:
            print(err)

    setup(**opts)

########################################################################################
# end block

ext_list = []

if 'develop' in sys.argv:
    ext_list = []

requirements = load_requirements('requirements.txt')
requirements_rpc = load_requirements('requirements-rpc.txt')

kwargs = dict(
    packages=find_packages(exclude=['tests', 'test_proj']+ext_list),
    ext_modules_list=ext_list,
    static_exclude_min=[
        'vstutils/templates/.*\.js$',
        'vstutils/static/bundle/.*\.js$'
    ],
    install_requires=[
        "django~=" + (os.environ.get('DJANGO_DEP', "") or "3.2.13"),
    ]
    + requirements
    + load_requirements('requirements-doc.txt'),
    extras_require={
        'test': load_requirements('requirements-test.txt'),
        'rpc': requirements_rpc,
        'ldap': load_requirements('requirements-ldap.txt'),
        'doc': ['django-docs==0.3.1'] + load_requirements('requirements-doc.txt'),
        'prod': load_requirements('requirements-prod.txt'),
        'stubs': load_requirements('requirements-stubs.txt'),
        'pil': ['Pillow~=8.4.0;python_version<"3.7"', 'Pillow~=9.0.0;python_version>"3.6"'],
        'boto3': [
            i.replace('libcloud', 'libcloud,boto3')
            for i in requirements
            if isinstance(i, str) and 'django-storages' in i
        ],
        'sqs': [
            i.replace('redis', 'sqs,redis')
            if isinstance(i, str) and 'celery' in i
            else i
            for i in requirements_rpc
        ],
        'librabbitmq': [
            i.replace('redis', 'librabbitmq,redis')
            if isinstance(i, str) and 'celery' in i
            else i
            for i in requirements_rpc
        ],
    },
    dependency_links=[
    ] + load_requirements('requirements-git.txt'),
    entry_points={
        'console_scripts': ['vstutilsctl=vstutils.__main__:cmd_execution']
    },
    project_urls={
        "Issue Tracker": "https://gitlab.com/vstconsulting/vstutils/issues",
        "Source Code": "https://gitlab.com/vstconsulting/vstutils",
        "Releases": "https://pypi.org/project/vstutils/#history",
        "Documentation": "https://vstutils.vstconsulting.net/",
    }
)

all_deps = []
for key, deps in kwargs['extras_require'].items():
    if key not in ('sqs', 'librabbitmq'):
        all_deps += deps

kwargs['extras_require']['all'] = all_deps

make_setup(**kwargs)



from django import template

from ..utils import translate

CONTEXT_VAR_NAME = '__lang__'
register = template.Library()


class TranslateTag(template.Node):
    def __init__(self, text, filters):
        self.text = text
        self.filters = filters

    def get_translate_callback(self, context):
        if CONTEXT_VAR_NAME in context:
            return context[CONTEXT_VAR_NAME].translate
        elif hasattr(context, 'request') \
            and hasattr(context.request, 'language') \
                and hasattr(context.request.language, 'translate'):
            return context.request.language.translate
        return translate

    def render(self, context):

        if str(self.text.var).startswith("$"):
            self.text.var = context[self.text.var[1:]]

        result = self.get_translate_callback(context)(self.text.var)
        if self.filters:
            tmpl = '{{ "' + result + '" |'
            tmpl += '|'.join(self.filters)
            tmpl += '}}'
            result = context.template.engine.from_string(tmpl).render(context=context)

        return result

    @classmethod
    def handle_token(cls, parser, token):
        """
        Class method to parse prefix node and return a Node.
        """
        bits = token.split_contents()

        if len(bits) < 2:
            raise template.TemplateSyntaxError(
                "'%s' takes at least one argument (phrase to translate)" % bits[0])  # nocv

        return cls(parser.compile_filter(bits[1]), bits[2:])


class TranslateErrorsTag(template.Node):
    def __init__(self, errors_list):
        self.errors_list = errors_list

    def render(self, context):
        trans = context.request.language.translate
        result = ''
        errors = self.errors_list.resolve(context)
        for error in errors.as_data():
            translated = trans(str(error.message))
            if error.params:
                translated = translated % {k: trans(v) for k, v in error.params.items()}
            result += f'<li>{translated}</li>'

        return f'<ul class="form-errors">{result}</ul>'


@register.tag('translate_text')
def translate_text(parser, token):
    """
    Translates text to language from request

    Usage::

        {% translate_text path [filters...] %}

    Examples::

        {% translate_text "username" capfirst cut:" "  %}
    """
    return TranslateTag.handle_token(parser, token)


@register.tag('translate_errors')
def translate_errors(parser, token):
    """
    Translates every error and its parameters from given ErrorList and renders it in ul list

    Usage:
        {% translate_errors form.non_field_errors %}
    """
    bits = token.split_contents()
    if len(bits) < 2:
        raise template.TemplateSyntaxError('translate_errors takes at least one argument')  # nocv
    return TranslateErrorsTag(template.Variable(bits[1]))



import json

from django import template
from django.templatetags.static import StaticNode

register = template.Library()


class StaticTag(StaticNode):
    def render(self, context):
        original_static = super().render(context)
        host = context.get('host_url', '')
        return (
            f'{host}{original_static}' if original_static else original_static
        )


@register.tag('static_path')
def do_static(parser, token):
    """
    Joins the given path with the STATIC_URL setting + host full path.

    Usage::

        {% static path [as varname] %}

    Examples::

        {% static "myapp/css/base.css" %}
        {% static variable_with_path %}
        {% static "myapp/css/base.css" as admin_base_css %}
        {% static variable_with_path as varname %}
    """
    return StaticTag.handle_token(parser, token)


@register.tag('static')
def do_static_origin(parser, token):
    """
    Joins the given path with the STATIC_URL setting + host full path.

    Usage::

        {% static path [as varname] %}

    Examples::

        {% static "myapp/css/base.css" %}
        {% static variable_with_path %}
        {% static "myapp/css/base.css" as admin_base_css %}
        {% static variable_with_path as varname %}
    """
    return StaticNode.handle_token(parser, token)  # nocv


@register.filter(is_safe=True)
def jsonify(object):
    return json.dumps(object)  # nocv



import json

from django.contrib import admin
from django.utils.html import format_html
from django.conf import settings
from django import forms

from . import models


class _BaseViewAdmin(admin.ModelAdmin):

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(models.TwoFactor)
class TwoFactorAdmin(_BaseViewAdmin):
    fields = list_display = ('pk', 'user')

    def user(self, instance):
        return getattr(instance.user, 'email', str(instance.user))

    def has_delete_permission(self, request, obj=None):
        return obj is not None and (not obj.user.is_superuser or obj.pk == request.user.pk)


@admin.register(models.Language)
class LanguageAdmin(_BaseViewAdmin):
    list_display = ('code', 'name')
    fields = list_display + ('translations',)

    def translations(self, instance):
        return format_html(
            '<pre>{}</pre>',
            json.dumps(instance.translations, indent=4)
        )


class CustomTranslationsForm(forms.ModelForm):
    code = forms.ChoiceField(choices=settings.LANGUAGES)


if settings.ENABLE_CUSTOM_TRANSLATIONS:
    @admin.register(models.CustomTranslations)
    class CustomTranslationsAdmin(admin.ModelAdmin):
        list_display = ('original', 'code', 'translated')
        form = CustomTranslationsForm



# Stubs for vstutils.api.responses (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by... hands in shell:
# from rest_framework import status
# template = '\nclass {name}(BaseResponseClass):\n    status_code = status.{name}\n\n\nclass Response{code}({name}):\n    ...\n\n'
# st = [(s, getattr(status, s)) for s in dir(status) if s.startswith('HTTP_')]
# for name, code in st: print(template.format(name=name, code=code), end='')

from rest_framework import response, status
from typing import Any, Dict, Text, SupportsFloat, Optional


class BaseResponseClass(response.Response):
    data: Any
    timings: Optional[Dict[Text, SupportsFloat]]
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...


class HTTP_100_CONTINUE(BaseResponseClass):
    status_code = status.HTTP_100_CONTINUE


class Response100(HTTP_100_CONTINUE):
    ...


class HTTP_101_SWITCHING_PROTOCOLS(BaseResponseClass):
    status_code = status.HTTP_101_SWITCHING_PROTOCOLS


class Response101(HTTP_101_SWITCHING_PROTOCOLS):
    ...


class HTTP_200_OK(BaseResponseClass):
    status_code = status.HTTP_200_OK


class Response200(HTTP_200_OK):
    ...


class HTTP_201_CREATED(BaseResponseClass):
    status_code = status.HTTP_201_CREATED


class Response201(HTTP_201_CREATED):
    ...


class HTTP_202_ACCEPTED(BaseResponseClass):
    status_code = status.HTTP_202_ACCEPTED


class Response202(HTTP_202_ACCEPTED):
    ...


class HTTP_203_NON_AUTHORITATIVE_INFORMATION(BaseResponseClass):
    status_code = status.HTTP_203_NON_AUTHORITATIVE_INFORMATION


class Response203(HTTP_203_NON_AUTHORITATIVE_INFORMATION):
    ...


class HTTP_204_NO_CONTENT(BaseResponseClass):
    status_code = status.HTTP_204_NO_CONTENT


class Response204(HTTP_204_NO_CONTENT):
    ...


class HTTP_205_RESET_CONTENT(BaseResponseClass):
    status_code = status.HTTP_205_RESET_CONTENT


class Response205(HTTP_205_RESET_CONTENT):
    ...


class HTTP_206_PARTIAL_CONTENT(BaseResponseClass):
    status_code = status.HTTP_206_PARTIAL_CONTENT


class Response206(HTTP_206_PARTIAL_CONTENT):
    ...


class HTTP_207_MULTI_STATUS(BaseResponseClass):
    status_code = status.HTTP_207_MULTI_STATUS


class Response207(HTTP_207_MULTI_STATUS):
    ...


class HTTP_208_ALREADY_REPORTED(BaseResponseClass):
    status_code = status.HTTP_208_ALREADY_REPORTED


class Response208(HTTP_208_ALREADY_REPORTED):
    ...


class HTTP_226_IM_USED(BaseResponseClass):
    status_code = status.HTTP_226_IM_USED


class Response226(HTTP_226_IM_USED):
    ...


class HTTP_300_MULTIPLE_CHOICES(BaseResponseClass):
    status_code = status.HTTP_300_MULTIPLE_CHOICES


class Response300(HTTP_300_MULTIPLE_CHOICES):
    ...


class HTTP_301_MOVED_PERMANENTLY(BaseResponseClass):
    status_code = status.HTTP_301_MOVED_PERMANENTLY


class Response301(HTTP_301_MOVED_PERMANENTLY):
    ...


class HTTP_302_FOUND(BaseResponseClass):
    status_code = status.HTTP_302_FOUND


class Response302(HTTP_302_FOUND):
    ...


class HTTP_303_SEE_OTHER(BaseResponseClass):
    status_code = status.HTTP_303_SEE_OTHER


class Response303(HTTP_303_SEE_OTHER):
    ...


class HTTP_304_NOT_MODIFIED(BaseResponseClass):
    status_code = status.HTTP_304_NOT_MODIFIED


class Response304(HTTP_304_NOT_MODIFIED):
    ...


class HTTP_305_USE_PROXY(BaseResponseClass):
    status_code = status.HTTP_305_USE_PROXY


class Response305(HTTP_305_USE_PROXY):
    ...


class HTTP_306_RESERVED(BaseResponseClass):
    status_code = status.HTTP_306_RESERVED


class Response306(HTTP_306_RESERVED):
    ...


class HTTP_307_TEMPORARY_REDIRECT(BaseResponseClass):
    status_code = status.HTTP_307_TEMPORARY_REDIRECT


class Response307(HTTP_307_TEMPORARY_REDIRECT):
    ...


class HTTP_308_PERMANENT_REDIRECT(BaseResponseClass):
    status_code = status.HTTP_308_PERMANENT_REDIRECT


class Response308(HTTP_308_PERMANENT_REDIRECT):
    ...


class HTTP_400_BAD_REQUEST(BaseResponseClass):
    status_code = status.HTTP_400_BAD_REQUEST


class Response400(HTTP_400_BAD_REQUEST):
    ...


class HTTP_401_UNAUTHORIZED(BaseResponseClass):
    status_code = status.HTTP_401_UNAUTHORIZED


class Response401(HTTP_401_UNAUTHORIZED):
    ...


class HTTP_402_PAYMENT_REQUIRED(BaseResponseClass):
    status_code = status.HTTP_402_PAYMENT_REQUIRED


class Response402(HTTP_402_PAYMENT_REQUIRED):
    ...


class HTTP_403_FORBIDDEN(BaseResponseClass):
    status_code = status.HTTP_403_FORBIDDEN


class Response403(HTTP_403_FORBIDDEN):
    ...


class HTTP_404_NOT_FOUND(BaseResponseClass):
    status_code = status.HTTP_404_NOT_FOUND


class Response404(HTTP_404_NOT_FOUND):
    ...


class HTTP_405_METHOD_NOT_ALLOWED(BaseResponseClass):
    status_code = status.HTTP_405_METHOD_NOT_ALLOWED


class Response405(HTTP_405_METHOD_NOT_ALLOWED):
    ...


class HTTP_406_NOT_ACCEPTABLE(BaseResponseClass):
    status_code = status.HTTP_406_NOT_ACCEPTABLE


class Response406(HTTP_406_NOT_ACCEPTABLE):
    ...


class HTTP_407_PROXY_AUTHENTICATION_REQUIRED(BaseResponseClass):
    status_code = status.HTTP_407_PROXY_AUTHENTICATION_REQUIRED


class Response407(HTTP_407_PROXY_AUTHENTICATION_REQUIRED):
    ...


class HTTP_408_REQUEST_TIMEOUT(BaseResponseClass):
    status_code = status.HTTP_408_REQUEST_TIMEOUT


class Response408(HTTP_408_REQUEST_TIMEOUT):
    ...


class HTTP_409_CONFLICT(BaseResponseClass):
    status_code = status.HTTP_409_CONFLICT


class Response409(HTTP_409_CONFLICT):
    ...


class HTTP_410_GONE(BaseResponseClass):
    status_code = status.HTTP_410_GONE


class Response410(HTTP_410_GONE):
    ...


class HTTP_411_LENGTH_REQUIRED(BaseResponseClass):
    status_code = status.HTTP_411_LENGTH_REQUIRED


class Response411(HTTP_411_LENGTH_REQUIRED):
    ...


class HTTP_412_PRECONDITION_FAILED(BaseResponseClass):
    status_code = status.HTTP_412_PRECONDITION_FAILED


class Response412(HTTP_412_PRECONDITION_FAILED):
    ...


class HTTP_413_REQUEST_ENTITY_TOO_LARGE(BaseResponseClass):
    status_code = status.HTTP_413_REQUEST_ENTITY_TOO_LARGE


class Response413(HTTP_413_REQUEST_ENTITY_TOO_LARGE):
    ...


class HTTP_414_REQUEST_URI_TOO_LONG(BaseResponseClass):
    status_code = status.HTTP_414_REQUEST_URI_TOO_LONG


class Response414(HTTP_414_REQUEST_URI_TOO_LONG):
    ...


class HTTP_415_UNSUPPORTED_MEDIA_TYPE(BaseResponseClass):
    status_code = status.HTTP_415_UNSUPPORTED_MEDIA_TYPE


class Response415(HTTP_415_UNSUPPORTED_MEDIA_TYPE):
    ...


class HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE(BaseResponseClass):
    status_code = status.HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE


class Response416(HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE):
    ...


class HTTP_417_EXPECTATION_FAILED(BaseResponseClass):
    status_code = status.HTTP_417_EXPECTATION_FAILED


class Response417(HTTP_417_EXPECTATION_FAILED):
    ...


class HTTP_418_IM_A_TEAPOT(BaseResponseClass):
    status_code = status.HTTP_418_IM_A_TEAPOT


class Response418(HTTP_418_IM_A_TEAPOT):
    ...


class HTTP_422_UNPROCESSABLE_ENTITY(BaseResponseClass):
    status_code = status.HTTP_422_UNPROCESSABLE_ENTITY


class Response422(HTTP_422_UNPROCESSABLE_ENTITY):
    ...


class HTTP_423_LOCKED(BaseResponseClass):
    status_code = status.HTTP_423_LOCKED


class Response423(HTTP_423_LOCKED):
    ...


class HTTP_424_FAILED_DEPENDENCY(BaseResponseClass):
    status_code = status.HTTP_424_FAILED_DEPENDENCY


class Response424(HTTP_424_FAILED_DEPENDENCY):
    ...


class HTTP_426_UPGRADE_REQUIRED(BaseResponseClass):
    status_code = status.HTTP_426_UPGRADE_REQUIRED


class Response426(HTTP_426_UPGRADE_REQUIRED):
    ...


class HTTP_428_PRECONDITION_REQUIRED(BaseResponseClass):
    status_code = status.HTTP_428_PRECONDITION_REQUIRED


class Response428(HTTP_428_PRECONDITION_REQUIRED):
    ...


class HTTP_429_TOO_MANY_REQUESTS(BaseResponseClass):
    status_code = status.HTTP_429_TOO_MANY_REQUESTS


class Response429(HTTP_429_TOO_MANY_REQUESTS):
    ...


class HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE(BaseResponseClass):
    status_code = status.HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE


class Response431(HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE):
    ...


class HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS(BaseResponseClass):
    status_code = status.HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS


class Response451(HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS):
    ...


class HTTP_500_INTERNAL_SERVER_ERROR(BaseResponseClass):
    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR


class Response500(HTTP_500_INTERNAL_SERVER_ERROR):
    ...


class HTTP_501_NOT_IMPLEMENTED(BaseResponseClass):
    status_code = status.HTTP_501_NOT_IMPLEMENTED


class Response501(HTTP_501_NOT_IMPLEMENTED):
    ...


class HTTP_502_BAD_GATEWAY(BaseResponseClass):
    status_code = status.HTTP_502_BAD_GATEWAY


class Response502(HTTP_502_BAD_GATEWAY):
    ...


class HTTP_503_SERVICE_UNAVAILABLE(BaseResponseClass):
    status_code = status.HTTP_503_SERVICE_UNAVAILABLE


class Response503(HTTP_503_SERVICE_UNAVAILABLE):
    ...


class HTTP_504_GATEWAY_TIMEOUT(BaseResponseClass):
    status_code = status.HTTP_504_GATEWAY_TIMEOUT


class Response504(HTTP_504_GATEWAY_TIMEOUT):
    ...


class HTTP_505_HTTP_VERSION_NOT_SUPPORTED(BaseResponseClass):
    status_code = status.HTTP_505_HTTP_VERSION_NOT_SUPPORTED


class Response505(HTTP_505_HTTP_VERSION_NOT_SUPPORTED):
    ...


class HTTP_506_VARIANT_ALSO_NEGOTIATES(BaseResponseClass):
    status_code = status.HTTP_506_VARIANT_ALSO_NEGOTIATES


class Response506(HTTP_506_VARIANT_ALSO_NEGOTIATES):
    ...


class HTTP_507_INSUFFICIENT_STORAGE(BaseResponseClass):
    status_code = status.HTTP_507_INSUFFICIENT_STORAGE


class Response507(HTTP_507_INSUFFICIENT_STORAGE):
    ...


class HTTP_508_LOOP_DETECTED(BaseResponseClass):
    status_code = status.HTTP_508_LOOP_DETECTED


class Response508(HTTP_508_LOOP_DETECTED):
    ...


class HTTP_509_BANDWIDTH_LIMIT_EXCEEDED(BaseResponseClass):
    status_code = status.HTTP_509_BANDWIDTH_LIMIT_EXCEEDED


class Response509(HTTP_509_BANDWIDTH_LIMIT_EXCEEDED):
    ...


class HTTP_510_NOT_EXTENDED(BaseResponseClass):
    status_code = status.HTTP_510_NOT_EXTENDED


class Response510(HTTP_510_NOT_EXTENDED):
    ...


class HTTP_511_NETWORK_AUTHENTICATION_REQUIRED(BaseResponseClass):
    status_code = status.HTTP_511_NETWORK_AUTHENTICATION_REQUIRED


class Response511(HTTP_511_NETWORK_AUTHENTICATION_REQUIRED):
    ...



import typing as _t
import os
import gc
import pwd
import sys
from tempfile import gettempdir

import environ
from django.contrib import admin
from django.utils.functional import lazy
from drf_yasg import errors
import rest_framework
import orjson
import ormsgpack

from configparserc import config as cconfig
from .tools import get_file_value

from . import __version__ as VSTUTILS_VERSION, __file__ as vstutils_file


SIMPLE_OBJECT_SETTINGS_TYPE = _t.Dict[_t.Text, _t.Dict[_t.Text, _t.Any]]


class Env(dict):
    __slots__ = ()

    def __getitem__(self, item):
        if item in self:
            return super().__getitem__(item)
        default = ''
        if ':-' in item:
            item, default = item.split(':-')
        return os.environ.get(item, default)


# MAIN Variables
##############################################################
interpreter_dir: _t.Text = os.path.dirname(sys.executable or 'python')
PYTHON_INTERPRETER: _t.Text = '/'.join([interpreter_dir, 'python'] if interpreter_dir else 'python')
VSTUTILS_DIR: _t.Text = os.path.dirname(os.path.abspath(vstutils_file))
VST_PROJECT: _t.Text = os.getenv("VST_PROJECT", "vstutils")
VST_PROJECT_LIB_NAME: _t.Text = os.getenv("VST_PROJECT_LIB_NAME", VST_PROJECT)
VST_PROJECT_LIB: _t.Text = os.getenv("VST_PROJECT_LIB", VST_PROJECT_LIB_NAME)
ENV_NAME: _t.Text = os.getenv("VST_PROJECT_ENV", VST_PROJECT_LIB.upper())
vst_project_module = __import__(VST_PROJECT)
vst_lib_module = __import__(VST_PROJECT_LIB_NAME) if VST_PROJECT != VST_PROJECT_LIB else vst_project_module
PROJECT_LIB_VERSION: _t.Text = getattr(vst_lib_module, '__version__', VSTUTILS_VERSION)
PROJECT_VERSION: _t.Text = getattr(vst_project_module, '__version__', PROJECT_LIB_VERSION)
FULL_VERSION: _t.Text = f'{PROJECT_VERSION}_{PROJECT_LIB_VERSION}_{VSTUTILS_VERSION}'
PROJECT_GUI_NAME: _t.Text = os.getenv("VST_PROJECT_GUI_NAME", ENV_NAME[0].upper()+ENV_NAME[1:].lower())

PY_VER: _t.SupportsInt = sys.version_info.major
TMP_DIR: _t.Text = gettempdir() or '/tmp'  # nosec
BASE_DIR: _t.Text = os.path.dirname(os.path.abspath(vst_lib_module.__file__))
VST_PROJECT_DIR: _t.Text = os.path.dirname(os.path.abspath(vst_project_module.__file__))
VST_PROJECT_LIB_DIR: _t.Text = os.path.dirname(os.path.abspath(vst_lib_module.__file__))
__kwargs: _t.Dict[_t.Text, _t.Any] = dict(
    PY=PY_VER,
    PY_VER='.'.join([str(i) for i in sys.version_info[:2]]),
    INTERPRETER=PYTHON_INTERPRETER,
    TMP=TMP_DIR,
    HOME=BASE_DIR,
    PROG=VST_PROJECT_DIR,
    LIB=BASE_DIR,
    VST=VSTUTILS_DIR,
    PROG_NAME=VST_PROJECT,
    LIB_NAME=VST_PROJECT_LIB_NAME,
    FULL_VERSION=FULL_VERSION,
    ENV=Env()
)
KWARGS: _t.Dict[_t.Text, _t.Any] = __kwargs

# Get settings from config
##############################################################
DEV_SETTINGS_FILE: _t.Text = os.getenv(
    f"{ENV_NAME}_DEV_SETTINGS_FILE",
    os.path.join(BASE_DIR, str(os.getenv("VST_DEV_SETTINGS")))
)
CONFIG_FILE: _t.Text = os.getenv(
    f"{ENV_NAME}_SETTINGS_FILE",
    f"/etc/{VST_PROJECT_LIB}/settings.ini"
)
PROJECT_CONFIG_FILE: _t.Text = os.getenv(
    f"{ENV_NAME}_LIB_SETTINGS_FILE",
    f"/etc/{VST_PROJECT}/settings.ini"
)
PROJECT_DEFAULTS_CONFIG = os.path.join(VST_PROJECT_LIB_DIR, 'settings.ini')
CONFIG_ENV_DATA_NAME: _t.Text = f"{ENV_NAME}_SETTINGS_DATA"

CONFIG_FILES = tuple(filter(bool, (
    PROJECT_DEFAULTS_CONFIG,
    '/etc/vstutils/settings.ini' if VST_PROJECT != 'test_proj' else None,
    '/etc/vstutils/settings.yml' if VST_PROJECT != 'test_proj' else None,
    os.path.splitext(CONFIG_FILE)[0] + '.yml' if CONFIG_FILE else None,
    CONFIG_FILE,
    os.path.splitext(PROJECT_CONFIG_FILE)[0] + '.yml' if PROJECT_CONFIG_FILE else None,
    PROJECT_CONFIG_FILE,
    os.path.splitext(DEV_SETTINGS_FILE)[0] + '.yml' if DEV_SETTINGS_FILE else None,
    DEV_SETTINGS_FILE,
)))

ConfigBoolType = cconfig.BoolType()
ConfigIntType = cconfig.IntType()
ConfigIntSecondsType = cconfig.IntSecondsType()
ConfigListType = cconfig.ListType()
ConfigStringType = cconfig.StrType()


class BoolOrStringType(cconfig.BaseType):
    def convert(self, value: _t.Any) -> _t.Optional[_t.Union[bool, _t.Text]]:
        if value is None or value == '':  # nocv
            return None
        elif value in ('false', 'False', False):
            return False
        return str(value)  # nocv


class FloatType(cconfig.BaseType):
    def convert(self, value: _t.Any) -> _t.SupportsFloat:
        return float(value)


class LocationsType(cconfig.ListType):
    def convert(self, value: _t.Any) -> _t.Union[_t.Iterable, _t.Text]:  # type: ignore
        result: _t.Iterable[_t.Text] = super().convert(value)
        if any(filter(lambda x: x.startswith('redis'), result)):  # type: ignore
            return result
        return value


class BackendSection(cconfig.Section):
    __slots__ = ()

    def key_handler_to_all(self, key):
        return super().key_handler_to_all(key).upper()


class BaseAppendSection(cconfig.AppendSection):
    __slots__ = ()
    get: _t.Callable[..., _t.Any]


class MainSection(BaseAppendSection):
    __slots__ = ()
    types_map = {
        'debug': ConfigBoolType,
        'enable_admin_panel': ConfigBoolType,
        'enable_registration': ConfigBoolType,
        'enable_custom_translations': ConfigBoolType,
        'allowed_hosts': cconfig.ListType(),
        'first_day_of_week': ConfigIntType,
        'enable_agreement_terms': ConfigBoolType,
        'agreement_terms_path': ConfigStringType,
        'enable_consent_to_processing': ConfigBoolType,
        'consent_to_processing_path': ConfigStringType,
    }


class WebSection(BaseAppendSection):
    __slots__ = ()
    types_map = {
        'allow_cors': ConfigBoolType,
        'cors_allowed_origins': cconfig.ListType(),
        'cors_allowed_origins_regexes': cconfig.ListType(),
        'cors_expose_headers': cconfig.ListType(),
        'cors_allow_methods': cconfig.ListType(),
        'cors_allow_headers': cconfig.ListType(),
        'cors_preflight_max_age': ConfigIntSecondsType,
        'session_timeout': ConfigIntSecondsType,
        'page_limit': ConfigIntType,
        'rest_page_limit': ConfigIntType,
        'public_openapi': ConfigBoolType,
        'openapi_cache_timeout': ConfigIntType,
        'enable_gravatar': ConfigBoolType,
        'rest_swagger': ConfigBoolType,
        'request_max_size': cconfig.BytesSizeType(),
        'x_frame_options': cconfig.StrType(),
        'use_x_forwarded_host': ConfigBoolType,
        'use_x_forwarded_port': ConfigBoolType,
        'password_reset_timeout_days': ConfigIntType,
        'secure_browser_xss_filter': ConfigBoolType,
        'secure_content_type_nosniff': ConfigBoolType,
        'secure_hsts_include_subdomains': ConfigBoolType,
        'secure_hsts_preload': ConfigBoolType,
        'secure_hsts_seconds': ConfigIntSecondsType,
        'health_throttle_rate': ConfigIntType,
        'bulk_threads': ConfigIntType,
        'max_tfa_attempts': ConfigIntType,
        'etag_default_timeout': ConfigIntSecondsType,
        'allow_auto_image_resize': ConfigBoolType,
    }


class DatabasesSection(BaseAppendSection):
    __slots__ = ()

    types_map = {
        'databases_without_cte_support': cconfig.ListType(),
    }


class DBSection(BackendSection):
    __slots__ = ()
    types_map = {
        'conn_max_age': ConfigIntSecondsType,
        'atomic_requests': ConfigBoolType,
        'autocommit': ConfigBoolType,
        'disable_server_side_cursors': ConfigBoolType,
    }


class DBTestSection(DBSection):
    __slots__ = ()
    type_serialize = ConfigBoolType
    type_create_db = ConfigBoolType
    type_create_user = ConfigBoolType
    type_dependencies = cconfig.ListType()


class DBOptionsSection(cconfig.Section):
    __slots__ = ()
    types_map = {
        'timeout': ConfigIntSecondsType,
        'connect_timeout': ConfigIntSecondsType,
        'read_timeout': ConfigIntSecondsType,
        'write_timeout': ConfigIntSecondsType,
        'isolation_level': ConfigIntType,
    }


class CacheSection(BackendSection):
    __slots__ = ()
    types_map = {
        'timeout': ConfigIntSecondsType,
        'location': LocationsType(),
    }


class CacheOptionsSection(BackendSection):
    types_map = {
        'binary': ConfigBoolType,
        'no_delay': ConfigBoolType,
        'ignore_exc': ConfigBoolType,
        'ignore_exceptions': ConfigBoolType,
        'use_pooling': ConfigBoolType,
        'close_connection': ConfigBoolType,
        'max_entries': ConfigIntType,
        'cull_frequency': ConfigIntType,
        'max_pool_size': ConfigIntType,
        'pickle_version': ConfigIntType,
        'socket_connect_timeout': ConfigIntSecondsType,
        'socket_timeout': ConfigIntSecondsType,
    }

    def key_handler_to_all(self, key):
        backend = self.parent['backend']
        if backend.split('.')[-1] in ('PyLibMCCache', 'PyMemcacheCache'):
            return key
        return super(CacheOptionsSection, self).key_handler_to_all(key)


class SentinelsSection(cconfig.Section):
    def all(self) -> _t.List[_t.Tuple[_t.Text, _t.SupportsInt]]:  # type: ignore
        return list((k, int(v)) for k, v in super().all().items())


class CachePoolKwargsSection(cconfig.Section):
    types_map = {
        'max_connections': ConfigIntType,
        'retry_on_timeout': ConfigBoolType,
        'skip_full_coverage_check': ConfigBoolType,
        'socket_keepalive': ConfigBoolType,
        'socket_connect_timeout': ConfigIntSecondsType,
        'socket_timeout': ConfigIntSecondsType,
    }


class CacheBehaviorsSection(cconfig.Section):
    types_map = {
        'ketama': ConfigBoolType,
        'no_block': ConfigBoolType,
        'tcp_nodelay': ConfigBoolType,
        'tcp_keepalive': ConfigBoolType,
        'connect_timeout': ConfigIntType,
        'send_timeout': ConfigIntType,
        'receive_timeout': ConfigIntType,
        '_poll_timeout': ConfigIntType,
        'remove_failed': ConfigIntType,
        'retry_timeout': ConfigIntType,
        'dead_timeout': ConfigIntType,
    }


class MailSection(BaseAppendSection):
    __slots__ = ()
    types_map = {
        'port': ConfigIntType,
        'tls': ConfigBoolType,
        'ssl': ConfigBoolType,
        'send_confirmation': ConfigBoolType,
        'authenticate_after_registration': ConfigBoolType,
        'agreement_terms': ConfigBoolType,
    }


class UWSGISection(cconfig.Section):
    __slots__ = ()
    type_daemon = ConfigBoolType


class RPCSection(BaseAppendSection):
    __slots__ = ()
    type_map = {
        'concurrency"': ConfigIntType,
        'prefetch_multiplier"': ConfigIntType,
        'max_tasks_per_child"': ConfigIntType,
        'heartbeat"': ConfigIntType,
        'results_expiry_days"': ConfigIntType,
        'create_instance_attempts"': ConfigIntType,
        'enable_worker"': ConfigBoolType
    }


class RPCBrokerSection(BaseAppendSection):
    __slots__ = ()
    types_map = {
        'visibility_timeout': ConfigIntSecondsType,
        'wait_time_seconds': ConfigIntSecondsType,
        'max_retries': ConfigIntType,
        'polling_interval': FloatType(),
    }


class WorkerSection(BaseAppendSection):
    __slots__ = ()
    types_map = {
        'beat': ConfigBoolType,
        'events': ConfigBoolType,
        'task-events': ConfigBoolType,
        'without-gossip': ConfigBoolType,
        'without-mingle': ConfigBoolType,
        'without-heartbeat': ConfigBoolType,
        'purge': ConfigBoolType,
        'discard': ConfigBoolType,
    }


class CentrifugoSection(cconfig.Section):
    __slots__ = ()
    type_address = cconfig.StrType()
    type_public_address = cconfig.StrType()
    type_api_key = cconfig.StrType()
    type_token_hmac_secret_key = cconfig.StrType()
    type_timeout = ConfigIntSecondsType
    type_verify = ConfigBoolType


class ThrottleSection(BaseAppendSection):
    __slots__ = ()
    types_map = {
        'rate': ConfigStringType,
        'actions': ConfigListType,
    }


class Boto3Subsection(BackendSection):
    types_map = {
        'querystring_auth': ConfigBoolType,
        'querystring_expire': ConfigIntSecondsType,
        'is_gzipped': ConfigBoolType,
        's3_use_ssl': ConfigBoolType,
        's3_secure_urls': ConfigBoolType,
        's3_verify': BoolOrStringType(),
        's3_max_memory_size': ConfigIntType,
        's3_file_overwrite': ConfigBoolType,
        'content_types': ConfigListType,
        'gzip_content_types': ConfigListType,
    }

    def key_handler_to_all(self, key):
        key_uppercase = super().key_handler_to_all(key).upper()
        if key_uppercase != 'GZIP_CONTENT_TYPES':
            return f'AWS_{key_uppercase}'
        return key_uppercase


env = environ.Env(
    # set casting, default value
    DEBUG=(bool, False),
    DJANGO_LOG_LEVEL=(str, 'WARNING'),
    TIMEZONE=(str, 'UTC'),
)

config: cconfig.ConfigParserC = cconfig.ConfigParserC(
    format_kwargs=KWARGS,
    section_defaults={
        'main': {
            'debug': env('DEBUG'),
            'allowed_hosts': ('*',),
            'timezone': env('TIMEZONE'),
            'first_day_of_week': 0,
            'log_level': env('DJANGO_LOG_LEVEL'),
            'enable_admin_panel': ConfigBoolType(os.getenv(f'{ENV_NAME}_ENABLE_ADMIN_PANEL', 'false')),
            'enable_registration': ConfigBoolType(os.getenv(f'{ENV_NAME}_ENABLE_REGISTRATION', 'false')),
            'enable_custom_translations': False,
            'ldap-server': None,
            'ldap-default-domain': '',
            'ldap-auth_format': 'cn=<username>,<domain>',
            'language_cookie_name': 'lang',
            'agreement_terms_path': f'/etc/{VST_PROJECT_LIB}/terms.md',
            'consent_to_processing_path': f'/etc/{VST_PROJECT_LIB}/consent_to_processing.md',
        },
        'web': {
            'allow_cors': False,
            'cors_allowed_origins': [],
            'cors_allowed_origins_regexes': [],
            'cors_expose_headers': [],
            'cors_preflight_max_age': '1d',
            'session_timeout': '2w',
            'static_files_url': '/static/',
            'page_limit': 20,
            'rest_page_limit': 1000,
            'rest_swagger_description': (vst_project_module.__doc__ or vst_lib_module.__doc__),
            'public_openapi': False,
            'openapi_cache_timeout': 120,
            'enable_gravatar': True,
            'request_max_size': 2621440,
            'x_frame_options': 'SAMEORIGIN',
            'use_x_forwarded_host': False,
            'use_x_forwarded_port': False,
            'password_reset_timeout_days': 1,
            'secure_browser_xss_filter': False,
            'secure_content_type_nosniff': False,
            'secure_hsts_include_subdomains': False,
            'secure_hsts_preload': False,
            'secure_hsts_seconds': 0,
            'health_throttle_rate': 60,
            'bulk_threads': 3,
            'max_tfa_attempts': ConfigIntType(os.getenv(f'{ENV_NAME}_MAX_TFA_ATTEMPTS', 5)),
            'etag_default_timeout': ConfigIntSecondsType(os.getenv(f'{ENV_NAME}_ETAG_TIMEOUT', '1d')),
            'allow_auto_image_resize': True,
        },
        'database': {
            **env.db(default=f'sqlite:///{KWARGS["PROG"]}/db.{KWARGS["PROG_NAME"]}.sqlite3'),
            'test': {
                'serialize': 'false'
            }
        },
        'databases': {
            "default_db": "default",
            "default_tablespace": "",
            "default_index_tablespace": "",
            "databases_without_cte_support": [],
            'default': {},
        },
        'cache': {
            **env.cache(
                default=f'filecache://{TMP_DIR}/{KWARGS["PROG_NAME"]}_django_cache_default_{KWARGS["PY_VER"]}'
            ),
            'timeout': '10m'
        },
        **{
            cache_name: {
                **env.cache(
                    var=f'{cache_name.upper()}_CACHE_URL',
                    default=f'filecache://{TMP_DIR}/{KWARGS["PROG_NAME"]}_django_cache_{cache_name}_{KWARGS["PY_VER"]}'
                ),
                'timeout': '10m'
            }
            for cache_name in ('locks', 'session', 'etag')
            if f'{cache_name.upper()}_CACHE_URL' in os.environ
        },
        'mail': {
            'port': 25,
            'user': "",
            'password': "",
            'tls': False,
            'host': None,
            'send_confirmation': os.getenv(f'{ENV_NAME}_SEND_CONFIRMATION_EMAIL', False),
            'send_email_retries': os.getenv(f'{ENV_NAME}_SEND_EMAIL_RETRIES', 3),
            'send_email_retry_delay': os.getenv(f'{ENV_NAME}_SEND_EMAIL_RETRY_DELAY', 10),
            'authenticate_after_registration': os.getenv(f'{ENV_NAME}_AUTHENTICATE_AFTER_REGISTRATION', False),
        },
        'contact': {
            'name': 'System Administrator'
        },
        'uwsgi': {
            'daemon': True
        },
        'rpc': {
            'concurrency': 4,
            'prefetch_multiplier': 1,
            'max_tasks_per_child': 1,
            'heartbeat': 10,
            'results_expiry_days': 1,
            'create_instance_attempts': 10,
            'default_delivery_mode': "persistent",
            'broker_transport_options': {}
        },
        'worker': {
            'app': os.getenv('VST_CELERY_APP', '{PROG_NAME}.wapp:app'),
            'loglevel': '{this[main][log_level]}',
            'pidfile': '/run/{PROG_NAME}_worker.pid',
            'autoscale': '{this[rpc][concurrency]},1',
            'hostname': f'{pwd.getpwuid(os.getuid()).pw_name}@%h',
            'beat': True
        },
        'storages': {
            'filesystem': {
                'media_url': '/media/',
                'media_root': os.getenv(f'{ENV_NAME}_MEDIA_ROOT', os.path.join(VST_PROJECT_DIR, "media"))
            }
        },
        'throttle': {
            'rate': os.getenv(f'{ENV_NAME}_GLOBAL_THROTTLE_RATE', ''),
            'actions': ConfigListType(os.getenv(f'{ENV_NAME}_GLOBAL_THROTTLE_ACTIONS', ('update', 'partial_update'))),
            'views': {},
        }
    },
    section_overload={
        'main': MainSection,
        'web': WebSection,
        'databases': DatabasesSection,
        'database': DBSection,
        'database.options': DBOptionsSection,
        'database.test': DBTestSection,
        'worker': WorkerSection,
        'cache': CacheSection,
        'cache.options': CacheOptionsSection,
        'cache.options.sentinels': SentinelsSection,
        'cache.options.connection_pool_kwargs': CachePoolKwargsSection,
        'cache.options.behaviors': CacheBehaviorsSection,
        'locks': CacheSection,
        'locks.options': CacheOptionsSection,
        'locks.options.sentinels': SentinelsSection,
        'locks.options.connection_pool_kwargs': CachePoolKwargsSection,
        'locks.options.behaviors': CacheBehaviorsSection,
        'session': CacheSection,
        'session.options': CacheOptionsSection,
        'session.options.sentinels': SentinelsSection,
        'session.options.connection_pool_kwargs': CachePoolKwargsSection,
        'session.options.behaviors': CacheBehaviorsSection,
        'etag': CacheSection,
        'etag.options': CacheOptionsSection,
        'etag.options.sentinels': SentinelsSection,
        'etag.options.connection_pool_kwargs': CachePoolKwargsSection,
        'etag.options.behaviors': CacheBehaviorsSection,
        'mail': MailSection,
        'uwsgi': UWSGISection,
        'rpc': RPCSection,
        'rpc.broker_transport_options': RPCBrokerSection,
        'centrifugo': CentrifugoSection,
        'throttle': ThrottleSection,
        'storages.boto3': Boto3Subsection,
    },
    format_exclude_sections=('uwsgi',)
)

config.parse_files(CONFIG_FILES)
config.parse_text(os.getenv(CONFIG_ENV_DATA_NAME, ''))

main: MainSection = config['main']
web: WebSection = config['web']

# Secret file with key for hashing passwords
SECRET_FILE = os.getenv(
    f"{ENV_NAME}_SECRET_FILE",
    f"/etc/{VST_PROJECT_LIB}/secret"
)


def secret_key():
    return get_file_value(
        SECRET_FILE, '*sg17)9wa_e+4$n%7n7r_(kqwlsc^^xdoc3&px$hs)sbz(-ml1'
    )


SECRET_KEY: _t.Text = lazy(secret_key, str)()

# Main settings
##############################################################
# SECURITY WARNING: don't run with debug turned on in production!
TESTS_RUN: bool = any([True for i in sys.argv if i in ['testserver', 'test']])
LOCALRUN: bool = any([True for i in sys.argv if i not in ['collectstatic', 'runserver']]) or TESTS_RUN
TESTSERVER_RUN: bool = 'testserver' in sys.argv
DEBUG: bool = ConfigBoolType(os.getenv('DJANGO_DEBUG', main["debug"]))
ALLOWED_HOSTS: _t.Iterable = main["allowed_hosts"]
SECURE_PROXY_SSL_HEADER: _t.Tuple[_t.Text, _t.Text] = (
    web.get('secure_proxy_ssl_header_name', fallback='HTTP_X_FORWARDED_PROTOCOL'),
    web.get('secure_proxy_ssl_header_value', fallback='https')
)

# Include some addons if packages exists in env
##############################################################
# :django_celery_beat:
has_django_celery_beat = False
try:
    import django_celery_beat
    has_django_celery_beat = True
except ImportError:  # nocv
    pass
RPC_ENABLED: bool = has_django_celery_beat

# :docs:
HAS_DOCS: bool = False
try:
    import docs
    HAS_DOCS = True
except ImportError:  # nocv
    pass


# Applications definition
##############################################################
INSTALLED_APPS: _t.List[_t.Text] = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'corsheaders',
]

if has_django_celery_beat:
    INSTALLED_APPS.append('django_celery_beat')

INSTALLED_APPS += [
    'crispy_forms',
    'rest_framework',
    'rest_framework.authtoken',
    'django_filters',
]

if HAS_DOCS:
    INSTALLED_APPS.append('docs')

INSTALLED_APPS += ['drf_yasg']

ADDONS: _t.List[_t.Text] = [
    'vstutils',
    'vstutils.api',
]

INSTALLED_APPS += ADDONS

# Additional middleware and auth
##############################################################
MIDDLEWARE: _t.List[_t.Text] = [
    'vstutils.middleware.FrontendChangesNotifications',
    'vstutils.middleware.ExecuteTimeHeadersMiddleware',
    'htmlmin.middleware.HtmlMinifyMiddleware',
    'htmlmin.middleware.MarkRequestMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'vstutils.middleware.LangMiddleware',
    'vstutils.middleware.TwoFaMiddleware',
]

EXCLUDE_FROM_MINIFYING = []

MIDDLEWARE_ENDPOINT_CONTROL = {
    'remove': [
        'corsheaders.middleware.CorsMiddleware',
        'htmlmin.middleware.HtmlMinifyMiddleware',
        'htmlmin.middleware.MarkRequestMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'vstutils.middleware.LangMiddleware',
        'vstutils.middleware.TwoFaMiddleware',
        'vstutils.middleware.FrontendChangesNotifications',
    ],
    'prepend': [
        'vstutils.api.endpoint.BulkMiddleware'
    ],
    'append': []
}

MAX_TFA_ATTEMPTS: int = web['max_tfa_attempts']

# Allow cross-domain access
CORS_ORIGIN_ALLOW_ALL: bool = web['allow_cors']
CORS_ALLOWED_ORIGINS: _t.Sequence[str] = web['cors_allowed_origins']
CORS_ALLOWED_ORIGIN_REGEXES: _t.Sequence[_t.Union[str, _t.Pattern[str]]] = web['cors_allowed_origins_regexes']
CORS_EXPOSE_HEADERS: _t.Sequence = web['cors_expose_headers']
CORS_PREFLIGHT_MAX_AGE: int = web['cors_preflight_max_age']
if 'cors_allow_methods' in web:
    CORS_ALLOW_METHODS = web['cors_allow_methods']
if 'cors_allow_headers' in web:
    CORS_ALLOW_HEADERS = web['cors_allow_headers']

LDAP_SERVER: _t.Optional[_t.Text] = main["ldap-server"]
LDAP_DOMAIN: _t.Optional[_t.Text] = main["ldap-default-domain"]
LDAP_FORMAT: _t.Text = main["ldap-auth_format"]

DEFAULT_AUTH_PLUGINS: SIMPLE_OBJECT_SETTINGS_TYPE = {
    'LDAP': {
        "BACKEND": "vstutils.auth.LdapBackend"
    },
    'DJANGO': {
        "BACKEND": "django.contrib.auth.backends.ModelBackend"
    },
}

DEFAULT_AUTH_PLUGIN_LIST: _t.Text = 'DJANGO,LDAP'


def get_plugins():
    plugins = {}
    for plugin_name in main.getlist('auth-plugins', fallback=DEFAULT_AUTH_PLUGIN_LIST):
        if plugin_name in DEFAULT_AUTH_PLUGINS:
            data = DEFAULT_AUTH_PLUGINS[plugin_name]
            name = plugin_name
        else:
            data = {"BACKEND": plugin_name}
            name = list(filter(bool, plugin_name.split('.')))[-1].lower().replace('backend', '')
        plugins[name] = data
    return plugins


AUTH_PLUGINS: SIMPLE_OBJECT_SETTINGS_TYPE = lazy(get_plugins, dict)()

AUTHENTICATION_BACKENDS: _t.List[_t.Text] = [
    'vstutils.auth.AuthPluginsBackend'
]
CACHE_AUTH_USER = main.getboolean('auth-cache-user', fallback=False)

# Sessions settings
# https://docs.djangoproject.com/en/3.2/ref/settings/#sessions
SESSION_COOKIE_AGE: int = web["session_timeout"]
SESSION_ENGINE: _t.Text = 'django.contrib.sessions.backends.cached_db'
SESSION_CACHE_ALIAS: _t.Text = 'session'
SESSION_COOKIE_DOMAIN: _t.Optional[_t.Text] = os.getenv(
    'DJANGO_SESSION_COOKIE_DOMAIN',
    _t.cast(_t.Any, web.get('session_cookie_domain', fallback=None))
)
SESSION_SERIALIZER = 'vstutils.session.YamlSessionSerializer'

CSRF_COOKIE_AGE: int = SESSION_COOKIE_AGE
CSRF_COOKIE_DOMAIN: _t.Optional[_t.Text] = SESSION_COOKIE_DOMAIN
CSRF_TRUSTED_ORIGINS: _t.List = web.getlist('csrf_trusted_origins', fallback=SESSION_COOKIE_DOMAIN)
CSRF_COOKIE_SECURE: bool = web.getboolean('csrf_cookie_secure', fallback=False)

SECURE_BROWSER_XSS_FILTER = web['secure_browser_xss_filter']
SECURE_CONTENT_TYPE_NOSNIFF = web['secure_content_type_nosniff']
SECURE_HSTS_INCLUDE_SUBDOMAINS = web['secure_hsts_include_subdomains']
SECURE_HSTS_PRELOAD = web['secure_hsts_preload']
SECURE_HSTS_SECONDS = web['secure_hsts_seconds']


# Password validation
# https://docs.djangoproject.com/en/3.2/ref/settings/#auth-password-validators
AUTH_PASSWORD_VALIDATORS: _t.List[_t.Dict] = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
        'OPTIONS': {
            'min_length': 0,
        },
    },
]
LOGIN_URL: _t.Text = '/login/'
LOGOUT_URL: _t.Text = '/logout/'
LOGIN_REDIRECT_URL: _t.Text = '/'

PASSWORD_RESET_TIMEOUT_DAYS: int = web['password_reset_timeout_days']

# Main controller settings
##############################################################
# Module with urls
ROOT_URLCONF: _t.Text = os.getenv('VST_ROOT_URLCONF', f'{VST_PROJECT}.urls')

# wsgi appilcation settings
WSGI: _t.Text = os.getenv('VST_WSGI', f'{VST_PROJECT}.wsgi')
WSGI_APPLICATION: _t.Text = f"{WSGI}.application"
UWSGI_APPLICATION: _t.Text = f'{WSGI}:application'

uwsgi_settings: cconfig.Section = config['uwsgi']
WEB_DAEMON = uwsgi_settings.getboolean('daemon', fallback=True)
WEB_DAEMON_LOGFILE: _t.Text = uwsgi_settings.get('log_file', fallback='/dev/null')  # type: ignore
WEB_ADDRPORT: _t.Text = uwsgi_settings.get('addrport', fallback=':8080')  # type: ignore

DATA_UPLOAD_MAX_MEMORY_SIZE = web['request_max_size']
X_FRAME_OPTIONS = web['x_frame_options']
USE_X_FORWARDED_HOST = web['use_x_forwarded_host']
USE_X_FORWARDED_PORT = web['use_x_forwarded_port']


# Templates settings
##############################################################
TEMPLATES: _t.List[_t.Dict] = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(VST_PROJECT_DIR, 'api/templates'),
            os.path.join(VST_PROJECT_DIR, 'gui/templates'),
            os.path.join(VST_PROJECT_DIR, 'templates'),
            os.path.join(BASE_DIR, 'api/templates'),
            os.path.join(BASE_DIR, 'gui/templates'),
            os.path.join(BASE_DIR, 'templates'),
            os.path.join(VSTUTILS_DIR, 'templates'),
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'vstutils.gui.context.settings_constants',
                'vstutils.gui.context.project_args',
                'vstutils.gui.context.pwa_context',
                'vstutils.gui.context.headers_context',
            ],
            'builtins': [
                'vstutils.templatetags.translation',
            ]
        },
    },
]


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.2/howto/static-files/
##############################################################
STATIC_URL: _t.Text = web["static_files_url"]

PROJECT_STATIC_DIRS = lazy(lambda: list(filter(bool, (
    os.path.join(VST_PROJECT_DIR, 'static'),
    os.path.join(BASE_DIR, 'static') if BASE_DIR != VST_PROJECT_DIR else None,
    os.path.join(VSTUTILS_DIR, 'static'),
))), tuple)()
STATIC_FILES_FOLDERS = lazy(lambda: list(filter(bool, (
    *PROJECT_STATIC_DIRS,
    os.path.join(os.path.dirname(admin.__file__), 'static'),
    os.path.join(os.path.dirname(errors.__file__), 'static'),
    os.path.join(os.path.dirname(rest_framework.__file__), 'static')
))), list)()

if LOCALRUN:
    STATICFILES_DIRS = list(STATIC_FILES_FOLDERS)

STATICFILES_FINDERS: _t.Tuple[_t.Text, _t.Text] = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)

if not LOCALRUN:
    STATIC_ROOT = os.path.join(VST_PROJECT_DIR, 'static')  # nocv


# Documentation files
# http://django-docs.readthedocs.io/en/latest/#docs-access-optional
##############################################################
DOCS_ROOT: _t.Text = os.path.join(BASE_DIR, 'doc/html')
DOCS_ACCESS: _t.Text = 'public'
DOC_URL: _t.Text = "/docs/"

if HAS_DOCS:
    EXCLUDE_FROM_MINIFYING.append(DOC_URL.lstrip('/'))


# Database settings.
# Read more: https://docs.djangoproject.com/en/3.2/ref/settings/#databases
##############################################################
DATABASES: SIMPLE_OBJECT_SETTINGS_TYPE


def parse_db(params):
    data: DBSection = config.get_section_instance('database')
    data.update({**config['database'].all(), **params[1].all()})
    data_all = data.all()
    USED_ENGINES.add(data_all.get('ENGINE'))
    return params[0], data_all


USED_ENGINES: _t.Set[_t.Text] = set()
DATABASES = dict(map(parse_db, filter(lambda x: isinstance(x[1], cconfig.Section), config['databases'].items())))

if 'django.db.backends.mysql' in USED_ENGINES:  # nocv
    try:
        import mysql
    except ImportError:
        try:
            import pymysql
            pymysql.install_as_MySQLdb()
        except ImportError:
            pass

for db in filter(lambda x: x.get('ENGINE', None) == 'django.db.backends.sqlite3', DATABASES.values()):
    try:
        db['OPTIONS'].setdefault('timeout', 20)
    except:  # nocv
        pass

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'
DEFAULT_TABLESPACE = config['databases'].get('default_tablespace', fallback='')
DEFAULT_INDEX_TABLESPACE = config['databases'].get('default_index_tablespace', fallback='')
DATABASES_WITHOUT_CTE_SUPPORT = config['databases'].get('databases_without_cte_support', fallback='')
# Cache settings.
# Read more: https://docs.djangoproject.com/en/3.2/ref/settings/#caches
##############################################################
default_cache = config['cache'].all()
session_cache = config['session'].all() or config['cache'].all()
session_cache['TIMEOUT'] = SESSION_COOKIE_AGE
etag_cache = config['etag'].all() or config['cache'].all()
etag_cache['TIMEOUT'] = web['etag_default_timeout']

CACHES: SIMPLE_OBJECT_SETTINGS_TYPE = {
    'default': default_cache,
    "locks": config['locks'].all() or default_cache,
    "session": session_cache,
    "etag": etag_cache,
}

if any([True for c in CACHES.values() if 'OPTIONS' in c and c['OPTIONS'].get('SENTINELS')]):
    DJANGO_REDIS_CONNECTION_FACTORY = 'django_redis.pool.SentinelConnectionFactory'


# E-Mail settings
# https://docs.djangoproject.com/en/3.2/ref/settings/#email-host
##############################################################
if 'EMAIL_URL' in os.environ:
    vars().update(env.email('EMAIL_URL'))  # nocv
else:
    mail = config['mail']
    EMAIL_BACKEND: _t.Text = 'django.core.mail.backends.smtp.EmailBackend'
    EMAIL_PORT = mail["port"]
    EMAIL_HOST_USER = mail["user"]
    EMAIL_HOST_PASSWORD = mail["password"]
    if mail.get('tls', None) is not None:
        EMAIL_USE_TLS = mail['tls']  # nocv
    if mail.get('ssl', None) is not None:
        EMAIL_USE_SSL = mail['ssl']  # nocv
    EMAIL_HOST = mail["host"]
    if EMAIL_HOST is None:
        EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'

EMAIL_FROM_ADDRESS = mail.get("from_address", EMAIL_HOST_USER)

SEND_CONFIRMATION_EMAIL: bool = mail["send_confirmation"]
SEND_EMAIL_RETRIES: int = mail['send_email_retries']
SEND_MESSAGE_RETRY_DELAY: int = mail['send_email_retry_delay']
AUTHENTICATE_AFTER_REGISTRATION: bool = mail["authenticate_after_registration"]
REGISTRATION_HASHERS: tuple = mail.getlist(
    'send_confirmation_uid_hashers',
    'django.contrib.auth.hashers.MD5PasswordHasher'
)


# API settings
##############################################################
VST_API_URL: _t.Text = os.getenv("VST_API_URL", "api")
VST_API_VERSION: _t.Text = os.getenv("VST_API_VERSION", r'v1')
API_URL: _t.Text = VST_API_URL
ENDPOINT_VIEW_CLASS: _t.Text = 'vstutils.api.endpoint.EndpointViewSet'
HAS_COREAPI: bool = False
API_CREATE_SWAGGER: bool = web.getboolean('rest_swagger', fallback=('drf_yasg' in INSTALLED_APPS))
SWAGGER_API_DESCRIPTION: _t.Text = web['rest_swagger_description']
TERMS_URL: _t.Text = ''
CONTACT: _t.Dict = config['contact'].all()
SCHEMA_CACHE_TIMEOUT = web['openapi_cache_timeout']
HEALTH_THROTTLE_RATE: _t.Text = f"{web['health_throttle_rate']}/minute"
OPENAPI_VIEW_CLASS: _t.Text = 'vstutils.api.schema.views.OpenApiView'
BULK_THREADS = web['bulk_threads']

OPENAPI_EXTRA_LINKS: SIMPLE_OBJECT_SETTINGS_TYPE = {
    'vstutils': {
        'url': 'https://github.com/vstconsulting/vstutils.git',
        'name': 'VST Utils sources'
    }
}

SWAGGER_SETTINGS: _t.Dict = {
    'DEFAULT_INFO': 'vstutils.api.schema.info.api_info',
    'DEFAULT_AUTO_SCHEMA_CLASS': 'vstutils.api.schema.schema.VSTAutoSchema',
    'DEFAULT_GENERATOR_CLASS': 'vstutils.api.schema.generators.VSTSchemaGenerator',
    'DEEP_LINKING': True,
    'SECURITY_DEFINITIONS': {
        'basic': {
            'type': 'basic'
        }
    },
    'LOGIN_URL': 'login',
    'LOGOUT_URL': 'logout',
}


API: SIMPLE_OBJECT_SETTINGS_TYPE = {
    VST_API_VERSION: {}
}

HEALTH_BACKEND_CLASS: _t.Text = 'vstutils.api.health.DefaultBackend'

# Rest Api settings
# http://www.django-rest-framework.org/api-guide/settings/
##############################################################
PAGE_LIMIT: int = web["page_limit"]
REST_FRAMEWORK: _t.Dict = {
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.SessionAuthentication',
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.TokenAuthentication',
    ),
    'DEFAULT_RENDERER_CLASSES': (
        "vstutils.api.renderers.ORJSONRenderer",
        "vstutils.api.renderers.MsgpackRenderer",
        'rest_framework.renderers.BrowsableAPIRenderer',
        'rest_framework.renderers.MultiPartRenderer',
    ),
    'DEFAULT_PARSER_CLASSES': (
        "drf_orjson_renderer.parsers.ORJSONParser",
        "vstutils.api.parsers.MsgpackParser",
        'rest_framework.parsers.FormParser',
        'rest_framework.parsers.MultiPartParser'
    ),
    'DEFAULT_PERMISSION_CLASSES': [
        'vstutils.api.permissions.IsAuthenticatedOpenApiRequest'
    ],
    "ORJSON_RENDERER_OPTIONS": (
        orjson.OPT_NON_STR_KEYS,
        orjson.OPT_SERIALIZE_DATACLASS,
        orjson.OPT_SERIALIZE_NUMPY,
        orjson.OPT_SERIALIZE_UUID,
    ),
    "MSGPACK_RENDERER_OPTIONS": (
        ormsgpack.OPT_NON_STR_KEYS,
        ormsgpack.OPT_SERIALIZE_NUMPY,
        ormsgpack.OPT_NAIVE_UTC,
    ),
    'EXCEPTION_HANDLER': 'vstutils.api.base.exception_handler',
    'DEFAULT_FILTER_BACKENDS': [
        'vstutils.api.filter_backends.DjangoFilterBackend',
        'vstutils.api.filter_backends.OrderingFilterBackend',
        'vstutils.api.filter_backends.HideHiddenFilterBackend',
        'vstutils.api.filter_backends.SelectRelatedFilterBackend',
        'rest_framework.filters.SearchFilter',
    ],
    'SEARCH_PARAM': '__search',
    'DEFAULT_PAGINATION_CLASS': 'vstutils.api.pagination.LimitOffsetPagination',
    'PAGE_SIZE': web["rest_page_limit"],
    'DEFAULT_SCHEMA_CLASS': 'vstutils.api.base.AutoSchema',
    'DEFAULT_METADATA_CLASS': 'vstutils.api.meta.VSTMetadata',
    'DEFAULT_VERSIONING_CLASS': 'rest_framework.versioning.NamespaceVersioning',
    'SCHEMA_COERCE_PATH_PK': False,
    'SCHEMA_COERCE_METHOD_NAMES': {
        'create': 'add',
        'list': 'list',
        'retrieve': 'get',
        'update': 'update',
        'partial_update': 'edit',
        'destroy': 'remove',
    },
    'DEFAULT_THROTTLE_CLASSES': (
        'vstutils.api.throttling.ActionBasedThrottle',
    )
}

OPTIMIZE_GET_BY_VALUES = True


# Internationalization
# https://docs.djangoproject.com/en/3.2/topics/i18n/
##############################################################
LANGUAGE_CODE: _t.Text = 'en'

LANGUAGES: _t.Tuple = (
    ('en', 'English'),
    ('ru', 'Русский'),
    ('cn', '简体中文'),
    ('vi', 'Tiếng Việt'),
)

TIME_ZONE: _t.Text = main["timezone"]
USE_I18N: bool = True
USE_L10N: bool = True
USE_TZ: bool = True
FIRST_DAY_OF_WEEK = main['first_day_of_week']
LANGUAGE_COOKIE_NAME = main['language_cookie_name']
ENABLE_CUSTOM_TRANSLATIONS = main['enable_custom_translations']


# LOGGING settings
##############################################################
LOG_LEVEL: _t.Text = os.getenv('DJANGO_LOG_LEVEL', main["log_level"])
LOG_LEVEL = os.getenv(f'{VST_PROJECT_LIB.upper()}_LOG_LEVEL', LOG_LEVEL).upper()
LOG_FORMAT: _t.Text = "[%(asctime)s] %(levelname)s [%(name)s.%(funcName)s:%(lineno)d] %(message)s"
LOG_DATE_FORMAT: _t.Text = "%d/%b/%Y %H:%M:%S"

default_logger_data = {
    'handlers': ['console', 'file'],
    'level': LOG_LEVEL,
    'propagate': True,
}

LOGGING: _t.Dict = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': LOG_FORMAT,
            'datefmt': LOG_DATE_FORMAT
        },
    },
    'handlers': {
        'console': {
            'level': LOG_LEVEL,
            'formatter': 'standard',
            'class': 'logging.StreamHandler',
            'stream': sys.stdout,
        },
        'file': {
            'level': LOG_LEVEL,
            'class': 'logging.FileHandler',
            'filename': WEB_DAEMON_LOGFILE
        } if WEB_DAEMON_LOGFILE != '/dev/null' else {'class': 'logging.NullHandler', 'level': LOG_LEVEL},
    },
    'loggers': {
        VST_PROJECT_LIB: default_logger_data,
        VST_PROJECT_LIB_NAME: default_logger_data,
        VST_PROJECT: default_logger_data,
        'vstutils': default_logger_data,
        'drf_yasg.generators': default_logger_data,
    }
}

if main.getboolean('enable_django_logs', fallback=False):  # nocv
    LOGGING['loggers']['django'] = default_logger_data

CASE_SENSITIVE_API_FILTER = web.getboolean('case_sensitive_api_filter', fallback=True)

SILENCED_SYSTEM_CHECKS: _t.List = [
    "urls.W005",
    "fields.W122"
]


# Celery broker settings
# Read more: http://docs.celeryproject.org/en/latest/userguide/configuration.html#conf-broker-settings
##############################################################
if RPC_ENABLED:
    rpc: RPCSection = config['rpc']
    CELERY_BROKER_TRANSPORT_OPTIONS: _t.Dict = rpc['broker_transport_options'].all()
    __broker_url = rpc.get("connection", fallback="file:///tmp")
    if __broker_url.startswith("file://"):
        __broker_folder = __broker_url.split("://", 1)[1]
        CELERY_BROKER_URL = "filesystem://"
        CELERY_BROKER_TRANSPORT_OPTIONS.update({
            "data_folder_in": __broker_folder,
            "data_folder_out": __broker_folder,
            "data_folder_processed": __broker_folder,
        })
        CELERY_RESULT_BACKEND = __broker_url
    else:  # nocv
        CELERY_BROKER_URL = __broker_url
        CELERY_RESULT_BACKEND = rpc.get("result_backend", fallback=CELERY_BROKER_URL)

    CELERY_WORKER_CONCURRENCY = rpc["concurrency"]
    CELERY_WORKER_HIJACK_ROOT_LOGGER = False
    CELERY_TASK_IGNORE_RESULT = True
    CELERYD_PREFETCH_MULTIPLIER = rpc["prefetch_multiplier"]
    CELERYD_MAX_TASKS_PER_CHILD = rpc["max_tasks_per_child"]
    CELERY_BROKER_HEARTBEAT = rpc["heartbeat"]
    CELERY_ACCEPT_CONTENT = ['pickle', 'json']
    CELERY_TASK_SERIALIZER = 'pickle'
    CELERY_RESULT_EXPIRES = rpc["results_expiry_days"]
    CELERY_DEFAULT_DELIVERY_MODE = rpc["default_delivery_mode"]
    CELERY_BEAT_SCHEDULER = 'vstutils.celery_beat_scheduler:SingletonDatabaseScheduler'
    CELERY_TASK_CREATE_MISSING_QUEUES = True
    CELERY_TIMEZONE = TIME_ZONE

    CREATE_INSTANCE_ATTEMPTS = rpc.getint("create_instance_attempts", fallback=10)
    CONCURRENCY = rpc["concurrency"]
    WORKER_QUEUES = ['celery']
    RUN_WORKER = rpc.getboolean('enable_worker', fallback=True)

    if RUN_WORKER:
        WORKER_OPTIONS = config['worker'].all()

# View settings
##############################################################
ENABLE_ADMIN_PANEL = main['enable_admin_panel']
MANIFEST_CLASS = 'vstutils.gui.pwa_manifest.PWAManifest'

VIEWS: SIMPLE_OBJECT_SETTINGS_TYPE = {
    "GUI": {
        "BACKEND": 'vstutils.gui.views.GUIView',
        "OPTIONS": {
            'name': 'gui'
        }
    },
    "MANIFEST": {
        "BACKEND": 'vstutils.gui.views.ManifestView',
        "OPTIONS": {
            'name': 'pwa_manifest'
        }
    },
    "SERVICE_WORKER": {
        "BACKEND": 'vstutils.gui.views.SWView',
        "OPTIONS": {
            'name': 'service_worker'
        }
    },
    "OFFLINE": {
        "BACKEND": 'vstutils.gui.views.OfflineView',
        "OPTIONS": {
            'name': 'offline_gui'
        }
    },
    "LOGIN": {
        "BACKEND": 'vstutils.gui.views.Login',
        "OPTIONS": {
            'name': 'login'
        }
    },
    "LOGOUT": {
        "BACKEND": 'vstutils.gui.views.Logout',
        "OPTIONS": {
            'view_args': [{'next_page': '/'}],
            'name': 'logout'
        }
    },
    "PASSWORD_RESET": {
        "BACKEND": 'django.contrib.auth.views.PasswordResetView',
        "OPTIONS": {
            'name': 'password_reset',
            'view_kwargs': {
                'from_email': EMAIL_FROM_ADDRESS
            }
        }
    },
    "PASSWORD_RESET_CONFIRM": {
        "BACKEND": 'django.contrib.auth.views.PasswordResetConfirmView',
        "OPTIONS": {
            'name': 'password_reset_confirm'
        }
    },
    "PASSWORD_RESET_DONE": {
        "BACKEND": 'django.contrib.auth.views.PasswordResetDoneView',
        "OPTIONS": {
            'name': 'password_reset_done'
        }
    },
    "PASSWORD_RESET_COMPLETE": {
        "BACKEND": 'django.contrib.auth.views.PasswordResetCompleteView',
        "OPTIONS": {
            'name': 'password_reset_complete'
        }
    },
    "USER_REGISTRATION": {
        "BACKEND": 'vstutils.gui.views.Registration',
        "OPTIONS": {
            'name': 'user_registration'
        }
    },
    "TERMS": {
        "BACKEND": 'vstutils.gui.views.TermsView',
        "OPTIONS": {
            'name': 'terms'
        }
    },
    "CONSENT_TO_PROCESSING": {
        "BACKEND": 'vstutils.gui.views.ConsentToProcessingView',
        "OPTIONS": {
            'name': 'consent_to_processing'
        }
    }
}

GUI_VIEWS: _t.Dict[_t.Text, _t.Union[_t.Text, _t.Dict]] = {
    r'^$': 'GUI',
    r'^manifest.json$': 'MANIFEST',
    r'^service-worker.js$': 'SERVICE_WORKER',
    r'^offline.html$': 'OFFLINE',
}

ACCOUNT_VIEWS: _t.Dict[_t.Text, _t.Union[_t.Text, _t.Dict]] = {
    'LOGIN_URL': 'LOGIN',
    'LOGOUT_URL': 'LOGOUT',
    r'^password_reset/$': 'PASSWORD_RESET',
    r'^password_reset_done/$': 'PASSWORD_RESET_DONE',
    r'^password_reset_complete/$': 'PASSWORD_RESET_COMPLETE',
    r'^password_reset_confirm/(?P<uidb64>.*)/(?P<token>.*)/$': 'PASSWORD_RESET_CONFIRM',
}


def get_accounts_views_mapping():
    mapping = {**ACCOUNT_VIEWS}
    if REGISTRATION_URL not in mapping and REGISTRATION_URL not in ACCOUNT_VIEWS and REGISTRATION_ENABLED:
        mapping[REGISTRATION_URL] = 'USER_REGISTRATION'

    if REGISTRATION_URL in mapping:
        if ENABLE_AGREEMENT_TERMS and AGREEMENT_TEMRS_URL:
            mapping[AGREEMENT_TEMRS_URL] = 'TERMS'

        if ENABLE_CONSENT_TO_PROCESSING and CONSENT_TO_PROCESSING_URL:
            mapping[CONSENT_TO_PROCESSING_URL] = 'CONSENT_TO_PROCESSING'

    return mapping


URLS = lazy(lambda: {**GUI_VIEWS}, dict)()
ACCOUNT_URLS = lazy(get_accounts_views_mapping, dict)()
ACCOUNT_URL = '^account/'

REGISTRATION_URL = r'^registration/$'
REGISTRATION_ENABLED = main['enable_registration']

AGREEMENT_TEMRS_URL = r'^terms/$'
ENABLE_AGREEMENT_TERMS: bool = main.getboolean('enable_agreement_terms', fallback=REGISTRATION_ENABLED)
AGREEMENT_TERMS_PATH: str = main['agreement_terms_path']

CONSENT_TO_PROCESSING_URL = r'^consent_to_processing/$'
ENABLE_CONSENT_TO_PROCESSING: bool = main.getboolean('enable_consent_to_processing', fallback=REGISTRATION_ENABLED)
CONSENT_TO_PROCESSING_PATH: str = main['consent_to_processing_path']


PROJECT_GUI_MENU: _t.List[_t.Dict] = [
    {
        'name': 'System',
        'span_class': 'fa fa-cog',
        'sublinks': [
            {
                'name': 'Users',
                'url': '/user',
                'span_class': 'fa fa-user',
            },
        ]
    },
]

SPA_STATIC: _t.List[_t.Dict] = []

SPA_STATIC_FILES_PROVIDERS = {
    'vstutils': {
        'BACKEND': 'vstutils.static_files.WebpackJsonStaticObjectHandler',
        'OPTIONS': {
            'priority': 1,
            'entrypoint_name': 'spa',
        }
    },
    'oldstyle': {
        'BACKEND': 'vstutils.static_files.SPAStaticObjectHandler',
    }
}

# Centrifugo settings
CENTRIFUGO_CLIENT_KWARGS = config['centrifugo'].all()
CENTRIFUGO_PUBLIC_HOST = CENTRIFUGO_CLIENT_KWARGS.pop('public_address', None) or CENTRIFUGO_CLIENT_KWARGS.get('address')
NOTIFICATOR_CLIENT_CLASS = web.get(
    'notificator_client_class',
    fallback='vstutils.models.cent_notify.Notificator'
)

THROTTLE = config['throttle'].all()


# Storage settings
def get_default_storage_class():
    if LIBCLOUD_PROVIDERS.get('default'):
        return 'storages.backends.apache_libcloud.LibCloudStorage'  # nocv
    elif all([i in globals() for i in ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY', 'AWS_STORAGE_BUCKET_NAME']]):
        return 'storages.backends.s3boto3.S3Boto3Storage'

    return 'django.core.files.storage.FileSystemStorage'  # nocv


storages = config['storages']
LIBCLOUD_PROVIDERS: _t.Dict[_t.Text, _t.Dict] = {}
MEDIA_ROOT = storages['filesystem']['media_root']
MEDIA_URL = storages['filesystem']['media_url']

if 'libcloud' in storages:

    LIBCLOUD_PROVIDERS = {
        store_name: store_settings
        for store_name, store_settings in storages['libcloud'].all().items()
        if isinstance(store_settings, dict)
    }

    if LIBCLOUD_PROVIDERS and 'default' not in LIBCLOUD_PROVIDERS:
        DEFAULT_LIBCLOUD_PROVIDER = next(iter(LIBCLOUD_PROVIDERS))

if 'boto3' in storages:
    globals().update( storages['boto3'].all())

DEFAULT_FILE_STORAGE = storages.get('default', fallback=get_default_storage_class())

DOCKERRUN_MIGRATE_LOCK_ID = config['docker'].get('migrate_lock_id', VST_PROJECT_LIB_NAME)

DOCKERRUN_MIGRATE_LOCK_TIMEOUT = config['docker'].getint('migrate_lock_timeout', 15)

# Test settings for speedup tests
##############################################################
if TESTS_RUN:
    gc.disable()
    gc.set_threshold(0)
    CELERY_TASK_ALWAYS_EAGER = True
    EMAIL_BACKEND = 'django.core.mail.backends.locmem.EmailBackend'
    PASSWORD_HASHERS = ['django.contrib.auth.hashers.MD5PasswordHasher', ]
    REGISTRATION_HASHERS = ('django.contrib.auth.hashers.MD5PasswordHasher',)
    CONTACT = dict(
        some_extra_url='https://pypi.org/project/vstutils/', **CONTACT
    )
    CELERY_RESULT_BACKEND = 'cache'
    CELERY_CACHE_BACKEND = 'memory'
    CACHES = {
        name: {'BACKEND': 'django.core.cache.backends.locmem.LocMemCache'}
        for name in CACHES
    }
    BULK_THREADS = 10
    SEND_EMAIL_RETRIES = 10
    DEFAULT_FILE_STORAGE = lazy(lambda: 'inmemorystorage.InMemoryStorage', str)()
    CENTRIFUGO_CLIENT_KWARGS = {}


if not TESTSERVER_RUN and TESTS_RUN:
    SESSION_ENGINE = 'django.contrib.sessions.backends.cache'

# User settings
##############################################################
ENABLE_GRAVATAR: bool = web["enable_gravatar"]

# Auto resize image settings
##############################################################
ALLOW_AUTO_IMAGE_RESIZE = web['allow_auto_image_resize']
